from __future__ import annotations

import math
import pathlib
import uuid
from typing import TYPE_CHECKING, Callable, Protocol

from negmas.common import MechanismState, NegotiatorMechanismInterface, TraceElement
from negmas.gb import ResponseType
from negmas.helpers.misc import make_callable
from negmas.helpers.strings import humanize_time
from negmas.negotiators import Negotiator
from negmas.outcomes.base_issue import Issue
from negmas.outcomes.common import Outcome, os_or_none
from negmas.outcomes.outcome_space import make_os
from negmas.outcomes.protocols import OutcomeSpace
from negmas.preferences import BaseUtilityFunction
from negmas.preferences.crisp_ufun import UtilityFunction
from negmas.preferences.ops import (
    kalai_points,
    max_relative_welfare_points,
    max_welfare_points,
    nash_points,
    pareto_frontier,
)

if TYPE_CHECKING:
    from matplotlib.axes import Axes

__all__ = [
    "plot_offer_utilities",
    "plot_mechanism_run",
    "plot_2dutils",
    "opacity_colorizer",
    "default_colorizer",
    "Colorizer",
]


Colorizer = Callable[[TraceElement], float]


def opacity_colorizer(t: TraceElement, alpha: float = 1.0):
    _ = t
    return alpha


def default_colorizer(t: TraceElement):
    if not t.responses:
        return 1.0
    return (
        0.9
        * len([_ for _ in t.responses.values() if _ == ResponseType.ACCEPT_OFFER])
        / len(t.responses)
        + 0.1
    )


def scatter_with_transparency(x, y, color, alpha_arr, ax, **kwarg):
    from matplotlib.colors import to_rgb  # , to_rgba

    r, g, b = to_rgb(color)
    # r, g, b, _ = to_rgba(color)
    color = [(r, g, b, alpha) for alpha in alpha_arr]
    ax.scatter(x, y, c=color, **kwarg)


def make_transparent(color, alpha):
    if alpha is None:
        return color
    from matplotlib.colors import to_rgb  # , to_rgba

    r, g, b = to_rgb(color)
    # r, g, b, _ = to_rgba(color)
    return (r, g, b, alpha)


def plot_with_trancparency(
    x,
    y,
    alpha,
    color,
    marker,
    ax,
    label,
    with_lines=False,
    alpha_global=None,
    linewidth: float | int = 1,
    linestyle="solid",
):
    # a.plot(
    #     xx,
    #     y,
    #     label=f"{name} ({i})",
    #     color=thecolor,
    #     linestyle="solid" if neg == negotiator else "dotted",
    #     linewidth=2 if neg == negotiator else 1,
    #     marker=None,
    # )
    if alpha_global is not None:
        alpha = [alpha_global * _ for _ in alpha]
    scatter_with_transparency(
        x,
        y,
        label=label,
        color=color,
        alpha_arr=alpha,
        ax=ax,
        marker=marker,
    )
    if with_lines:
        ax.plot(
            x,
            y,
            color=make_transparent(color, alpha_global),
            marker=None,
            linewidth=linewidth,
            linestyle=linestyle,
        )


ALL_MARKERS = ["s", "o", "v", "^", "<", ">", "p", "P", "h", "H", "1", "2", "3", "4"]
PROPOSALS_ALPHA = 0.9
AGREEMENT_ALPHA = 0.9
NASH_ALPHA = 0.6
KALAI_ALPHA = 0.6
RESERVED_ALPHA = 0.08
WELFARE_ALPHA = 0.6
AGREEMENT_SCALE = 10
NASH_SCALE = 4
KALAI_SCALE = 4
WELFARE_SCALE = 2
OUTCOMES_SCALE = 0.5
PARETO_SCALE = 1.5


class PlottableMechanism(Protocol):
    @property
    def outcome_space(self) -> OutcomeSpace:
        ...

    @property
    def negotiators(self) -> list[Negotiator]:
        ...

    @property
    def negotiator_ids(self) -> list[str]:
        ...

    @property
    def negotiator_names(self) -> list[str]:
        ...

    @property
    def nmi(self) -> NegotiatorMechanismInterface:
        ...

    @property
    def agreement(self) -> Outcome | None:
        ...

    @property
    def state(self) -> MechanismState:
        ...

    @property
    def full_trace(self) -> list[TraceElement]:
        ...

    def discrete_outcomes(self) -> list[Outcome]:
        ...

    def negotiator_index(self, source: str) -> int | None:
        ...


DEFAULT_COLORMAP = "jet"


def get_cmap(n, name=DEFAULT_COLORMAP):
    """Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
    RGB color; the keyword argument name must be a standard mpl colormap name."""
    import matplotlib.pyplot as plt

    return plt.cm.get_cmap(name, n)


def make_colors_and_markers(colors, markers, n: int, colormap=DEFAULT_COLORMAP):
    if not colors:
        cmap = get_cmap(n, colormap)
        colors = [cmap(i) for i in range(n)]
    if not markers:
        markers = [ALL_MARKERS[i % len(ALL_MARKERS)] for i in range(n)]
    return colors, markers


def plot_offer_utilities(
    trace: list[TraceElement],
    negotiator: str,
    plotting_ufuns: list[BaseUtilityFunction | None],
    plotting_negotiators: list[str],
    ignore_none_offers: bool = True,
    name_map: dict[str, str] | Callable[[str], str] | None = None,
    colors: list | None = None,
    markers: list | None = None,
    colormap: str = DEFAULT_COLORMAP,
    ax: Axes | None = None,  # type: ignore
    sharey=False,
    xdim: str = "relative_time",
    ylimits: tuple[float, float] | None = None,
    show_legend=True,
    show_x_label=True,
    ignore_markers_limit=200,
    show_reserved=True,
    colorizer: Colorizer | None = None,
    first_color_index: int = 0,
):
    import matplotlib.pyplot as plt

    if colorizer is None:
        colorizer = default_colorizer

    map_ = make_callable(name_map)
    if ax is None:
        _, ax = plt.subplots()  # type: ignore
    ax: Axes
    one_y = True
    axes = [ax] * len(plotting_negotiators)
    if not sharey and len(plotting_negotiators) == 2:
        axes = [ax, ax.twinx()]
        one_y = False

    colors, markers = make_colors_and_markers(
        colors, markers, len(plotting_negotiators), colormap
    )
    if first_color_index:
        colors = (
            colors[:first_color_index]
            + colors[first_color_index + 1 :]
            + [colors[first_color_index]]
        )
        markers = (
            markers[:first_color_index]
            + markers[first_color_index + 1 :]
            + [markers[first_color_index]]
        )

    if xdim.startswith("step") or xdim.startswith("round"):
        trace_info = [(_, _.step) for _ in trace if _.negotiator == negotiator]
    elif xdim.startswith("time") or xdim.startswith("real"):
        trace_info = [(_, _.time) for _ in trace if _.negotiator == negotiator]
    else:
        trace_info = [(_, _.relative_time) for _ in trace if _.negotiator == negotiator]
    x = [_[-1] for _ in trace_info]
    simple_offers_view = (
        len(plotting_negotiators) == 1 and plotting_negotiators[0] == negotiator
    )
    # reorder to plot the negotiator last if it was in the plotting-negotiators
    # if negotiator in plotting_negotiators:
    #     indx = plotting_negotiators.index(negotiator)
    #     plotting_negotiators = (
    #         plotting_negotiators[:indx]
    #         + plotting_negotiators[indx + 1 :]
    #         + [plotting_negotiators[indx]]
    #     )
    #     axes = axes[:indx] + axes[indx + 1 :] + [axes[indx]]
    #     plotting_ufuns = (
    #         plotting_ufuns[:indx] + plotting_ufuns[indx + 1 :] + [plotting_ufuns[indx]]
    #     )
    #     markers = markers[:indx] + markers[indx + 1 :] + [markers[indx]]
    #     colors = colors[:indx] + colors[indx + 1 :] + [colors[indx]]
    for i, (u, neg, a) in enumerate(zip(plotting_ufuns, plotting_negotiators, axes)):
        if u is None:
            continue
        name = map_(neg)
        alphas = [colorizer(_[0]) for _ in trace_info]
        y = [u(_[0].offer) for _ in trace_info]
        reserved = None
        if show_reserved:
            r = u.reserved_value
            if r is not None and math.isfinite(r):
                reserved = [r] * len(y)
        if not ignore_none_offers:
            xx, aa = x, alphas
        else:
            good_indices = [i for i, _ in enumerate(y) if _ is not None]
            xx = [x[_] for _ in good_indices]
            aa = [alphas[_] for _ in good_indices]
            y = [y[_] for _ in good_indices]
        thecolor = colors[i % len(colors)]
        # a.plot(
        #     xx,
        #     y,
        #     label=f"{name} ({i})",
        #     color=thecolor,
        #     linestyle="solid" if neg == negotiator else "dotted",
        #     linewidth=2 if neg == negotiator else 1,
        #     marker=None,
        # )
        plot_with_trancparency(
            x=xx,
            y=y,
            alpha=aa,
            color=thecolor,
            marker=markers[i % len(markers)]
            if len(trace_info) < ignore_markers_limit and neg == negotiator
            else "",
            label=f"Utility" if simple_offers_view else name,
            ax=a,
            with_lines=True,
            linestyle="solid" if neg == negotiator else ":",
            linewidth=1 if neg == negotiator else 0.5,
        )
        if reserved:
            a.plot(
                xx,
                reserved,
                # label=f"{name} ({i})",
                color=colors[i % len(colors)],
                linestyle="solid" if neg == negotiator else "dotted",
                linewidth=0.5,
            )
        if ylimits is not None:
            a.set_ylim(ylimits)
        a.set_ylabel(f"{name} ({i})" if not one_y or simple_offers_view else "utility")
        if show_legend and len(plotting_negotiators) == 2:
            a.legend(
                loc=f"upper {'left' if not i else 'right'}",
                bbox_to_anchor=(0.0, 1.4, 1.0, 0.12),
            )
    axes[0].set_title(f"{map_(negotiator)} Offers")
    if show_legend and len(plotting_negotiators) != 2:
        ax.legend(
            bbox_to_anchor=(0.0, 1.22, 1.0, 0.102),
            loc="upper center",
            ncol=2,
            mode="expand",
            borderaxespad=0.0,
            labelcolor=colors,
            # labelcolor="linecolor",
            # draggablebool=True,
        )
    if show_x_label:
        ax.set_xlabel(xdim)

    plt.tight_layout(w_pad=1.08, h_pad=1.12)


def plot_2dutils(
    trace: list[TraceElement],
    plotting_ufuns: list[UtilityFunction] | tuple[UtilityFunction],
    plotting_negotiators: list[str],
    offering_negotiators: list[str] | None = None,
    agreement: Outcome | None = None,
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
    with_lines: bool = True,
    show_annotations: bool = True,
    show_agreement: bool = False,
    show_pareto_distance: bool = True,
    show_nash_distance: bool = True,
    show_kalai_distance: bool = True,
    show_max_welfare_distance: bool = True,
    show_max_relative_welfare_distance: bool = True,
    show_reserved: bool = True,
    show_total_time=True,
    show_relative_time=True,
    show_n_steps=True,
    end_reason: str | None = None,
    extra_annotation: str | None = None,
    name_map: dict[str, str] | Callable[[str], str] | None = None,
    colors: list | None = None,
    markers: list[str] | None = None,
    colormap: str = DEFAULT_COLORMAP,
    ax: Axes | None = None,  # type: ignore
    colorizer: Colorizer | None = None,
    fast: bool = False,
):
    import matplotlib.patches as mpatches
    import matplotlib.pyplot as plt

    if not colorizer:
        colorizer = default_colorizer

    if ax is None:
        _, ax = plt.subplots()  # type: ignore
    ax: Axes
    map_ = make_callable(name_map)
    if not outcomes:
        outcome_space = os_or_none(outcome_space, issues, outcomes)
        if outcome_space:
            outcomes = list(outcome_space.enumerate_or_sample(10, 1000))
    if not outcomes:
        outcomes = list({_.offer for _ in trace})
    if not outcome_space:
        outcome_space = make_os(issues=issues, outcomes=outcomes)
    if not offering_negotiators:
        offering_negotiators = list({_.negotiator for _ in trace})

    utils = [tuple(f(o) for f in plotting_ufuns) for o in outcomes]
    yrange = max(_[1] for _ in utils) - min(_[1] for _ in utils)
    colors, markers = make_colors_and_markers(
        colors, markers, len(offering_negotiators), colormap
    )

    agreement_utility = tuple(u(agreement) for u in plotting_ufuns)
    unknown_agreement_utility = None in agreement_utility
    if unknown_agreement_utility:
        show_pareto_distance = show_nash_distance = False
    default_marker_size = plt.rcParams.get("lines.markersize", 20) ** 2
    ax.scatter(
        [_[0] for _ in utils],
        [_[1] for _ in utils],
        color="gray",
        marker=".",
        s=int(default_marker_size * OUTCOMES_SCALE),
    )
    agent_names = [map_(_) for _ in plotting_negotiators]
    if fast:
        frontier, frontier_outcome = [], []
        frontier_indices = []
        frontier_outcome = [frontier_outcome[i] for i in frontier_indices]
        nash_pts = []
        kalai_pts = []
        mwelfare_pts = []
        mrwelfare_pts = []
    else:
        frontier, frontier_outcome = pareto_frontier(
            ufuns=plotting_ufuns,
            issues=outcome_space.issues,  # type: ignore
            outcomes=outcomes if not issues else None,  # type: ignore
            sort_by_welfare=True,
        )
        frontier_indices = [
            i
            for i, _ in enumerate(frontier)
            if _[0] is not None
            and _[0] > float("-inf")
            and _[1] is not None
            and _[1] > float("-inf")
        ]
        frontier = [frontier[i] for i in frontier_indices]
        frontier_outcome = [frontier_outcome[i] for i in frontier_indices]

        nash_pts = nash_points(plotting_ufuns, frontier, outcome_space=outcome_space)
        kalai_pts = kalai_points(plotting_ufuns, frontier, outcome_space=outcome_space)
        mwelfare_pts = max_welfare_points(
            plotting_ufuns, frontier, outcome_space=outcome_space
        )
        mrwelfare_pts = max_relative_welfare_points(
            plotting_ufuns, frontier, outcome_space=outcome_space
        )

    if not nash_pts:
        show_nash_distance = False
    if not kalai_pts:
        show_kalai_distance = False
    if not mwelfare_pts:
        show_max_welfare_distance = False
    if not mrwelfare_pts:
        show_max_relative_welfare_distance = False
    pareto_distance = float("inf")
    nash_distance, kalai_distance = float("inf"), float("inf")
    max_welfare_distance, max_relative_welfare_distance = float("inf"), float("inf")
    f1, f2 = [_[0] for _ in frontier], [_[1] for _ in frontier]
    ax.scatter(
        f1, f2, color="red", marker="x", s=int(default_marker_size * PARETO_SCALE)
    )
    ax.set_xlabel(agent_names[0] + f"(0) utility")
    ax.set_ylabel(agent_names[1] + f"(1) utility")
    cu = agreement_utility
    if not unknown_agreement_utility and not fast:
        for nash, _ in nash_pts:
            nash_distance = min(
                nash_distance,
                math.sqrt((nash[0] - cu[0]) ** 2 + (nash[1] - cu[1]) ** 2),
            )
        for kalai, _ in kalai_pts:
            kalai_distance = min(
                kalai_distance,
                math.sqrt((kalai[0] - cu[0]) ** 2 + (kalai[1] - cu[1]) ** 2),
            )
        for pt, _ in mwelfare_pts:
            max_welfare_distance = min(
                max_welfare_distance,
                math.sqrt((pt[0] - cu[0]) ** 2 + (pt[1] - cu[1]) ** 2),
            )
        for pt, _ in mrwelfare_pts:
            max_relative_welfare_distance = min(
                max_relative_welfare_distance,
                math.sqrt((pt[0] - cu[0]) ** 2 + (pt[1] - cu[1]) ** 2),
            )
        for pu in frontier:
            dist = math.sqrt((pu[0] - cu[0]) ** 2 + (pu[1] - cu[1]) ** 2)
            if dist < pareto_distance:
                pareto_distance = dist
    if trace:
        n_steps = trace[-1].step + 1
        relative_time = trace[-1].relative_time
        total_time = trace[-1].time
    else:
        n_steps = relative_time = total_time = 0
    txt = ""
    if show_agreement:
        txt += f"Agreement:{agreement}\n"
    if not fast and show_pareto_distance and agreement is not None:
        txt += f"Pareto-distance={pareto_distance:5.2}\n"
    if not fast and show_nash_distance and agreement is not None:
        txt += f"Nash-distance={nash_distance:5.2}\n"
    if not fast and show_kalai_distance and agreement is not None:
        txt += f"Kalai-distance={kalai_distance:5.2}\n"
    if not fast and show_max_welfare_distance and agreement is not None:
        txt += f"MaxWelfare-distance={max_welfare_distance:5.2}\n"
    if not fast and show_max_relative_welfare_distance and agreement is not None:
        txt += f"MaxRelativeWelfare-distance={max_relative_welfare_distance:5.2}\n"
    if show_relative_time and relative_time:
        txt += f"Relative Time={relative_time:5.2}\n"
    if show_total_time:
        txt += f"Total Time={humanize_time(total_time, show_ms=True)}\n"
    if show_n_steps:
        txt += f"N. Steps={n_steps}\n"
    if end_reason:
        txt += f"{end_reason}\n"
    if extra_annotation:
        txt += f"{extra_annotation}\n"

    ax.text(
        0.05,
        0.05,
        txt,
        verticalalignment="bottom",
        transform=ax.transAxes,
        weight="bold",
    )

    if show_reserved:
        ranges = [
            plotting_ufuns[_].minmax(outcome_space=outcome_space)
            for _ in range(len(plotting_ufuns))
        ]
        for i, (mn, mx) in enumerate(ranges):
            if any(_ is None or not math.isfinite(_) for _ in (mn, mx)):
                x = []
                for a, neg in enumerate(offering_negotiators):
                    negtrace = [_ for _ in trace if _.negotiator == neg]
                    x += [plotting_ufuns[i](_.offer) for _ in negtrace]
                ranges[i] = (min(x), max(x))
        for i, (mn, mx) in enumerate(ranges):
            r = plotting_ufuns[i].reserved_value
            if r is None or not math.isfinite(r):
                r = mn
            if i == 0:
                pt = (r, ranges[1 - i][0])
                width = mx - r
                height = ranges[1 - i][1] - ranges[1 - i][0]
            else:
                pt = (ranges[1 - i][0], r)
                height = mx - r
                width = ranges[1 - i][1] - ranges[1 - i][0]

            ax.add_patch(
                mpatches.Rectangle(
                    pt,
                    width=width,
                    height=height,
                    fill=True,
                    color=colors[i % len(colors)],
                    alpha=RESERVED_ALPHA,
                    linewidth=0,
                )
            )
    for a, neg in enumerate(offering_negotiators):
        negtrace = [_ for _ in trace if _.negotiator == neg]
        x = [plotting_ufuns[0](_.offer) for _ in negtrace]
        y = [plotting_ufuns[1](_.offer) for _ in negtrace]
        alphas = [colorizer(_) for _ in negtrace]
        # (ax.scatter if not with_lines else ax.plot)(
        plot_with_trancparency(
            x=x,
            y=y,
            alpha=alphas,
            color=colors[a % len(colors)],
            alpha_global=PROPOSALS_ALPHA,
            label=f"{map_(neg)}",
            ax=ax,
            marker=markers[a % len(markers)],
            with_lines=with_lines,
            linestyle=":",
        )
    # if not fast and frontier:
    #     welfare, mx = [frontier[0]], sum(frontier[0])
    #     for u in frontier[1:]:
    #         if sum(u) < mx - 1e-12:
    #             break
    #         welfare.append(u)
    #     ax.scatter(
    #         [_[0] for _ in welfare],
    #         [_[1] for _ in welfare],
    #         color="magenta",
    #         label=f"Max. Welfare",
    #         marker="s",
    #         alpha=WELFARE_ALPHA,
    #         s=int(default_marker_size * WELFARE_SCALE),
    #     )
    #     if show_annotations:
    #         for f in welfare:
    #             ax.annotate(
    #                 "Max. Welfare",
    #                 xy=f,  # theta, radius
    #                 xytext=(
    #                     f[0] + 0.02,
    #                     f[1] + 0.02 * yrange,
    #                 ),
    #                 horizontalalignment="left",
    #                 verticalalignment="bottom",
    #             )
    #
    if agreement is not None:
        ax.scatter(
            [plotting_ufuns[0](agreement)],
            [plotting_ufuns[1](agreement)],
            color="black",
            marker="*",
            s=int(default_marker_size * AGREEMENT_SCALE),
            alpha=AGREEMENT_ALPHA,
            label="Agreement",
        )
        if show_annotations:
            ax.annotate(
                "Agreement",
                xy=agreement_utility,  # theta, radius
                xytext=(
                    agreement_utility[0] + 0.02,
                    agreement_utility[1] + 0.02,
                ),  # fraction, fraction
                horizontalalignment="left",
                verticalalignment="bottom",
            )
    if not fast:
        if kalai_pts:
            ax.scatter(
                [kalai[0] for kalai, _ in kalai_pts],
                [kalai[1] for kalai, _ in kalai_pts],
                color="green",
                label=f"Kalai Point",
                marker="P",
                alpha=KALAI_ALPHA,
                s=int(default_marker_size * KALAI_SCALE),
            )
            if show_annotations:
                for kalai, _ in kalai_pts:
                    ax.annotate(
                        "Kalai Point",
                        xy=kalai,  # theta, radius
                        xytext=(
                            kalai[0] + 0.02,
                            kalai[1] - 0.02 * yrange,
                        ),  # fraction, fraction
                        horizontalalignment="left",
                        verticalalignment="bottom",
                    )
        # if mrwelfare_pts:
        #     ax.scatter(
        #         [mrwelfare[0] for mrwelfare, _ in mrwelfare_pts],
        #         [mrwelfare[1] for mrwelfare, _ in mrwelfare_pts],
        #         color="cyan",
        #         label=f"Max Relative Welfare Points",
        #         marker="X",
        #         alpha=NASH_ALPHA,
        #         s=int(default_marker_size * NASH_SCALE),
        #     )
        #     if show_annotations:
        #         for mrwelfare, _ in mrwelfare_pts:
        #             ax.annotate(
        #                 "Max Relative Welfare Point",
        #                 xy=mrwelfare,  # theta, radius
        #                 xytext=(
        #                     mrwelfare[0] + 0.02,
        #                     mrwelfare[1] - 0.02 * yrange,
        #                 ),  # fraction, fraction
        #                 horizontalalignment="left",
        #                 verticalalignment="bottom",
        #             )

        if nash_pts:
            ax.scatter(
                [nash[0] for nash, _ in nash_pts],
                [nash[1] for nash, _ in nash_pts],
                color="brown",
                label=f"Nash Point",
                marker="^",
                alpha=NASH_ALPHA,
                s=int(default_marker_size * NASH_SCALE),
            )
            if show_annotations:
                for nash, _ in nash_pts:
                    ax.annotate(
                        "Nash Point",
                        xy=nash,  # theta, radius
                        xytext=(
                            nash[0] + 0.02,
                            nash[1] - 0.02 * yrange,
                        ),  # fraction, fraction
                        horizontalalignment="left",
                        verticalalignment="bottom",
                    )

        if mwelfare_pts:
            ax.scatter(
                [mwelfare[0] for mwelfare, _ in mwelfare_pts],
                [mwelfare[1] for mwelfare, _ in mwelfare_pts],
                color="brown",
                label=f"Max Welfare Points",
                marker="p",
                alpha=NASH_ALPHA,
                s=int(default_marker_size * NASH_SCALE),
            )
            if show_annotations:
                for mwelfare, _ in mwelfare_pts:
                    ax.annotate(
                        "Max Welfare Point",
                        xy=mwelfare,  # theta, radius
                        xytext=(
                            mwelfare[0] + 0.02,
                            mwelfare[1] - 0.02 * yrange,
                        ),  # fraction, fraction
                        horizontalalignment="left",
                        verticalalignment="bottom",
                    )
        ax.legend(
            bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
            loc="lower left",
            ncol=2,
            mode="expand",
            borderaxespad=0.0,
        )


def plot_mechanism_run(
    mechanism: PlottableMechanism,
    negotiators: tuple[int, int] | tuple[str, str] | None = (0, 1),
    save_fig: bool = False,
    path: str | None = None,
    fig_name: str | None = None,
    ignore_none_offers: bool = True,
    with_lines: bool = True,
    show_agreement: bool = False,
    show_pareto_distance: bool = True,
    show_nash_distance: bool = True,
    show_kalai_distance: bool = True,
    show_max_welfare_distance: bool = True,
    show_max_relative_welfare_distance: bool = False,
    show_end_reason: bool = True,
    show_annotations: bool = False,
    show_reserved: bool = True,
    show_total_time=True,
    show_relative_time=True,
    show_n_steps=True,
    colors: list | None = None,
    markers: list[str] | None = None,
    colormap: str = DEFAULT_COLORMAP,
    ylimits: tuple[float, float] | None = None,
    common_legend=True,
    extra_annotation: str = "",
    xdim: str = "relative_time",
    colorizer: Colorizer | None = None,
    only2d: bool = False,
    fast: bool = False,
    simple_offers_view: bool = False,
):
    import matplotlib.gridspec as gridspec
    import matplotlib.pyplot as plt

    if negotiators is None:
        negotiators = (0, 1)
    if len(negotiators) != 2:
        raise ValueError(
            f"Cannot plot the 2D plot for the mechanism run without knowing two plotting negotiators"
        )
    plotting_negotiators = []
    plotting_ufuns = []
    for n in negotiators:
        if isinstance(n, int):
            i = n
        else:
            i = mechanism.negotiator_index(n)
            if i is None:
                raise ValueError(f"Cannot find negotiator with ID {n}")
        plotting_negotiators.append(mechanism.negotiators[i].id)
        plotting_ufuns.append(mechanism.negotiators[i].ufun)

    name_map = dict(zip(mechanism.negotiator_ids, mechanism.negotiator_names))
    fig = plt.figure(figsize=(12.8, 4.8) if not only2d else None)
    if only2d:
        axu = fig.subplots()
    else:
        gs = gridspec.GridSpec(mechanism.nmi.n_negotiators, 2)
        axs = []
        colors, markers = make_colors_and_markers(
            colors, markers, len(mechanism.negotiators), colormap
        )

        all_ufuns = [_.ufun for _ in mechanism.negotiators]
        for a, neg in enumerate(mechanism.negotiator_ids):
            if a == 0:
                axs.append(fig.add_subplot(gs[a, 1]))
            else:
                axs.append(fig.add_subplot(gs[a, 1], sharex=axs[0]))
            plot_offer_utilities(
                trace=mechanism.full_trace,
                negotiator=neg,
                plotting_ufuns=[all_ufuns[a]] if simple_offers_view else all_ufuns,
                plotting_negotiators=[neg]
                if simple_offers_view
                else mechanism.negotiator_ids,
                ax=axs[-1],
                name_map=name_map,
                colors=colors,
                markers=markers,
                ignore_none_offers=ignore_none_offers,
                ylimits=ylimits,
                show_legend=(not common_legend or a == 0) and not simple_offers_view,
                show_x_label=a == len(mechanism.negotiator_ids) - 1,
                show_reserved=show_reserved,
                xdim=xdim,
                colorizer=colorizer,
                first_color_index=a if simple_offers_view else 0,
            )
        axu = fig.add_subplot(gs[:, 0])
    agreement = mechanism.agreement
    state = mechanism.state
    if not show_end_reason:
        reason = None
    else:
        if state.timedout:
            reason = "Negotiation Timedout"
        elif agreement is not None:
            reason = "Negotiation Success"
        elif state.has_error:
            reason = "Negotiation ERROR"
        elif agreement is not None:
            reason = "Agreemend Reached"
        elif state.broken:
            reason = "Negotiation Ended"
        elif agreement is None:
            reason = "No Agreement"
        else:
            reason = "Unknown state!!"

    plot_2dutils(
        trace=mechanism.full_trace,
        plotting_ufuns=plotting_ufuns,
        plotting_negotiators=plotting_negotiators,
        offering_negotiators=mechanism.negotiator_ids,
        outcome_space=mechanism.outcome_space,
        outcomes=mechanism.discrete_outcomes(),
        ax=axu,
        name_map=name_map,
        with_lines=with_lines,
        show_agreement=show_agreement,
        show_pareto_distance=show_pareto_distance,
        show_nash_distance=show_nash_distance,
        show_kalai_distance=show_kalai_distance,
        show_max_welfare_distance=show_max_welfare_distance,
        show_max_relative_welfare_distance=show_max_relative_welfare_distance,
        show_annotations=show_annotations,
        show_reserved=show_reserved,
        colors=colors,
        markers=markers,
        agreement=mechanism.agreement,
        end_reason=reason,
        extra_annotation=extra_annotation,
        colorizer=colorizer,
        show_total_time=show_total_time,
        show_relative_time=show_relative_time,
        show_n_steps=show_n_steps,
        fast=fast,
    )
    if save_fig:
        if fig_name is None:
            fig_name = str(uuid.uuid4()) + ".png"
        if path is None:
            path_ = pathlib.Path().absolute()
        else:
            path_ = pathlib.Path(path)
        path_.mkdir(parents=True, exist_ok=True)
        fig.savefig(str(path_ / fig_name), bbox_inches="tight", transparent=False, pad_inches=0.05)  # type: ignore
    return fig



from __future__ import annotations

import random
from random import choice

import hypothesis.strategies as st
from hypothesis import example, given, settings
from pytest import mark

import negmas
from negmas import (
    PolyAspiration,
    PresortingInverseUtilityFunction,
    all_negotiator_types,
)
from negmas.common import PreferencesChangeType
from negmas.gb.common import ResponseType
from negmas.gb.negotiators.timebased import AspirationNegotiator
from negmas.outcomes import Issue, make_issue
from negmas.outcomes.outcome_space import make_os
from negmas.preferences import LinearAdditiveUtilityFunction
from negmas.preferences import LinearAdditiveUtilityFunction as LUFun
from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun
from negmas.sao import EndImmediately, NoneOfferingPolicy, RejectAlways, SAOMechanism
from negmas.sao.common import SAOResponse, SAOState
from negmas.sao.negotiators.base import SAONegotiator
from negmas.sao.negotiators.modular.boa import make_boa

NEGTYPES = all_negotiator_types()


class SmartAspirationNegotiator(SAONegotiator):
    _inv = None  # The ufun invertor (finds outcomes in a utility range)
    _partner_first = None  # The best offer of the partner (assumed best for it)
    _min = None  # The minimum of my utility function
    _max = None  # The maximum of my utility function
    _best = None  # The best outcome for me

    def __init__(self, *args, **kwargs):
        # initialize the base SAONegoiator (MUST be done)
        super().__init__(*args, **kwargs)

        # Initialize the aspiration mixin to start at 1.0 and concede slowly
        self._asp = PolyAspiration(1.0, "boulware")

    def on_preferences_changed(self, changes):
        # create an initiaze an invertor for my ufun
        changes = [_ for _ in changes if _.type not in (PreferencesChangeType.Scale,)]
        if not changes:
            return
        self._inv = PresortingInverseUtilityFunction(self.ufun)  # type: ignore
        self._inv.init()

        # find worst and best outcomes for me
        worest, self._best = self.ufun.extreme_outcomes()  # type: ignore

        # and the correponding utility values
        self._min, self._max = self.ufun(worest), self.ufun(self._best)  # type: ignore

        # MUST call parent to avoid being called again for no reason
        super().on_preferences_changed(changes)

    def respond(self, state, source: str):
        offer = state.current_offer
        if offer is None:
            return ResponseType.REJECT_OFFER
        # set the partner's first offer when I receive it
        if not self._partner_first:
            self._partner_first = offer

        # accept if the offer is not worse for me than what I would have offered
        return super().respond(state, source)

    def propose(self, state):
        # calculate my current aspiration level (utility level at which I will offer and accept)
        a = (self._max - self._min) * self._asp.utility_at(  # type: ignore
            state.relative_time
        ) + self._min

        # find some outcomes (all if the outcome space is  discrete) above the aspiration level
        outcomes = self._inv.some((a - 1e-6, self._max + 1e-6), False)  # type: ignore
        # If there are no outcomes above the aspiration level, offer my best outcome
        if not outcomes:
            return self._best

        # else if I did not  recieve anything from the partner, offer any outcome above the aspiration level
        if not self._partner_first:
            return choice(outcomes)

        # otherwise, offer the outcome most similar to the partner's first offer (above the aspiration level)
        nearest, ndist = None, float("inf")
        for o in outcomes:
            d = sum((a - b) * (a - b) for a, b in zip(o, self._partner_first))
            if d < ndist:
                nearest, ndist = o, d
        return nearest


def try_negotiator(cls, replace_buyer=True, replace_seller=True, n_steps=100):
    buyer_cls = cls if replace_buyer else AspirationNegotiator
    seller_cls = cls if replace_seller else AspirationNegotiator

    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=10),
    ]

    # create the mechanism
    session = SAOMechanism(issues=issues, n_steps=n_steps)

    # define ufuns
    seller_utility = LUFun(
        values={  # type: ignore
            "price": IdentityFun(),
            "quantity": LinearFun(0.2),
            "delivery_time": AffineFun(-1, bias=9),
        },
        weights={"price": 1.0, "quantity": 1.0, "delivery_time": 10.0},
        outcome_space=session.outcome_space,
        reserved_value=15.0,
    ).scale_max(1.0)
    buyer_utility = LUFun(
        values={  # type: ignore
            "price": AffineFun(-1, bias=9.0),
            "quantity": LinearFun(0.2),
            "delivery_time": IdentityFun(),
        },
        outcome_space=session.outcome_space,
        reserved_value=10.0,
    ).scale_max(1.0)

    session.add(buyer_cls(name="buyer"), ufun=buyer_utility)  # type: ignore
    session.add(seller_cls(name="seller"), ufun=seller_utility)  # type: ignore
    session.run()
    return session


@given(
    opp=st.sampled_from(NEGTYPES),
    start=st.booleans(),
    rejector=st.sampled_from([EndImmediately, RejectAlways]),
)
@example(
    opp=negmas.sao.negotiators.timebased.AdditiveFirstFollowingTBNegotiator,
    start=True,
    rejector=negmas.sao.components.acceptance.EndImmediately,
)
@settings(deadline=500000)
def test_do_nothing_never_gets_agreements(opp, start, rejector):
    agent = make_boa(acceptance=rejector(), offering=NoneOfferingPolicy())
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(10, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    ufuns = [
        LinearAdditiveUtilityFunction.random(issues=issues),
        LinearAdditiveUtilityFunction.random(issues=issues),
    ]
    session = SAOMechanism(n_steps=1000, issues=issues)
    negs = [opp(), agent] if not start else [agent, opp()]
    for n, u in zip(negs, ufuns):
        session.add(n, preferences=u)

    assert session.run().agreement is None


@mark.parametrize(
    ["factory", "name", "short_name"], [(make_boa, "BOANegotiator", "BOA")]
)
def test_has_correct_type_name(factory, name, short_name):
    x = factory()
    assert x.type_name == name
    assert x.short_type_name == short_name


@mark.repeat(3)
def test_pend_works():
    os = make_os(
        [
            make_issue(10, "price"),
            make_issue(10, "quantity"),
            make_issue(["red", "green", "blue"], "color"),
        ]
    )
    for _ in range(50):
        ufuns = [
            LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
            LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
        ]
        n = 1000
        f = 0.01
        session = SAOMechanism(
            n_steps=None, time_limit=None, pend=f / n, outcome_space=os
        )
        for i, u in enumerate(ufuns):
            neg = AspirationNegotiator()
            assert session.add(neg, preferences=u)  # type: ignore
            assert len(session.negotiators) == (i + 1)

        assert abs(session.expected_relative_time - (f / (n + 1))) < 1e-8
        assert session.expected_remaining_time is None
        assert session.expected_remaining_steps is not None
        assert abs(session.expected_remaining_steps - n / f) < 4
        assert abs(session.relative_time - (f / (n + 1))) < 1e-8
        assert session.remaining_steps is None
        assert session.remaining_time is None
        assert session.state.step <= 10000 * n
        assert not session.state.started
        agreement = session.run().agreement
        assert session.state.started and session.state.ended
        if agreement is not None:
            break
    else:
        raise AssertionError(f"agreement failed in all runs")


def test_pend_per_second_works():
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(10, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    ufuns = [
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
    ]
    n = 10
    session = SAOMechanism(
        n_steps=None, time_limit=None, pend_per_second=1 / n, issues=issues
    )
    for u in ufuns:
        session.add(AspirationNegotiator(), preferences=u)  # type: ignore

    assert session.expected_relative_time < 1e-8
    assert (
        session.expected_remaining_time is not None
        and abs(session.expected_remaining_time - n) < 1e-8
    )
    assert session.expected_remaining_steps is None
    assert session.relative_time < 1e-8
    assert session.remaining_steps is None
    assert session.remaining_time is None
    session.run()
    assert session.state.time <= 100 * n


@mark.parametrize("s", [1, 3, 10, 101, 1000])
def test_nsteps_apply_as_round(s):
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(10, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    ufuns = [
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
    ]
    session = SAOMechanism(n_steps=s, issues=issues)
    for u in ufuns:
        assert session.add(AspirationNegotiator(), preferences=u)  # type: ignore

    assert session.expected_remaining_steps == s
    assert session.remaining_steps == s
    assert session.current_step == 0
    assert abs(session.relative_time - (1.0 / (s + 1))) < 1e-6
    assert session.remaining_time is None
    session.step()
    assert session.current_step == 1
    assert session.expected_remaining_steps == (s - 1)
    assert session.remaining_steps == s - 1
    assert abs(session.relative_time - (2.0 / (s + 1))) < 1e-6
    assert session.remaining_time is None
    session.run()
    ndone = session.current_step
    for nid in session.negotiator_ids:
        assert len(session.negotiator_offers(nid)) in (ndone, ndone - 1)
    assert session.state.step <= s


@mark.parametrize("s", [1, 3, 10, 101, 1000])
def test_nsteps_apply_as_step(s):
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(10, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    ufuns = [
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(issues=issues, reserved_value=0.0),
    ]
    session = SAOMechanism(n_steps=s, issues=issues, one_offer_per_step=True)
    for u in ufuns:
        assert session.add(AspirationNegotiator(), preferences=u)  # type: ignore

    assert session.expected_remaining_steps == s
    assert session.remaining_steps == s
    assert session.current_step == 0
    assert abs(session.relative_time - (1.0 / (s + 1))) < 1e-6
    assert session.remaining_time is None
    session.step()
    assert session.current_step == 1
    assert session.expected_remaining_steps == (s - 1)
    assert session.remaining_steps == s - 1
    assert abs(session.relative_time - (2.0 / (s + 1))) < 1e-6
    assert session.remaining_time is None
    session.run()
    ndone = session.current_step
    for nid in session.negotiator_ids:
        assert len(session.negotiator_offers(nid)) in (
            int(ndone / 2),
            int(ndone / 2) + 1,
            int((ndone - 1) / 2),
            int((ndone - 1) / 2) + 1,
        )
    assert session.state.step <= s


def test_basic_sao():
    n_steps = 100
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(5, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    os = make_os(issues)
    ufuns = [
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
    ]
    session = SAOMechanism(n_steps=n_steps, outcome_space=os, one_offer_per_step=True)
    agents = [AspirationNegotiator() for _ in range(len(ufuns))]
    for u, a in zip(ufuns, agents):
        assert session.add(a, ufun=u)  # type: ignore
    # offers = [os.random_outcome() for _ in range(n_steps)]
    assert session.expected_remaining_steps == n_steps
    assert session.remaining_steps == n_steps
    assert session.current_step == 0
    assert abs(session.relative_time - (1.0 / (n_steps + 1))) < 1e-6
    assert session.remaining_time is None
    assert not session.state.started and not session.state.running
    for i in range(n_steps):
        if not session.step().running:
            break
        assert (
            session.state.started and session.state.running
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            session.current_step == i + 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert session.expected_remaining_steps == (
            n_steps - i - 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            session.remaining_steps == n_steps - i - 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            abs(session.relative_time - ((i + 2) / (n_steps + 1))) < 1e-6
        ), f"{session.state=}\n{session.extended_trace=}"
        assert session.remaining_time is None
    assert session.state.started and not session.state.running
    assert session.state.step <= n_steps


def test_basic_sao_with_action():
    n_steps = 50
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(5, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    os = make_os(issues)
    ufuns = [
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
        LinearAdditiveUtilityFunction.random(outcome_space=os, reserved_value=0.0),
    ]
    session = SAOMechanism(n_steps=n_steps, outcome_space=os, one_offer_per_step=True)
    agents = [AspirationNegotiator() for _ in range(len(ufuns))]
    ids = [_.id for _ in agents]
    for u, a in zip(ufuns, agents):
        assert session.add(a, ufun=u)  # type: ignore
    offers = [os.random_outcome() for _ in range(n_steps)]
    assert session.expected_remaining_steps == n_steps
    assert session.remaining_steps == n_steps
    assert session.current_step == 0
    assert abs(session.relative_time - (1.0 / (n_steps + 1))) < 1e-6
    assert session.remaining_time is None
    assert not session.state.started and not session.state.running
    for i in range(n_steps):
        action = None
        pass_action = random.random() < 0.5
        if pass_action:
            ids = session.next_negotitor_ids()
            assert len(ids) == 1
            action = {ids[0]: SAOResponse(ResponseType.REJECT_OFFER, offers[i])}
        if not session.step(action).running:
            break
        if pass_action:
            state: SAOState = session.state  # type: ignore
            assert state.current_offer == offers[i]
            assert state.current_proposer == ids[0]
        assert (
            session.state.started and session.state.running
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            session.current_step == i + 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert session.expected_remaining_steps == (
            n_steps - i - 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            session.remaining_steps == n_steps - i - 1
        ), f"{session.state=}\n{session.extended_trace=}"
        assert (
            abs(session.relative_time - ((i + 2) / (n_steps + 1))) < 1e-6
        ), f"{session.state=}\n{session.extended_trace=}"
        assert session.remaining_time is None
    assert session.state.started and (
        not session.state.running or session.state.step >= n_steps
    ), f"Did not finish running:\n{session.extended_trace}"
    assert (
        session.state.step <= n_steps
    ), f"Ran for too long {session.state.step} but max expected is {n_steps} steps:\n{session.extended_trace}"


def test_smart_asipration():
    state = try_negotiator(SmartAspirationNegotiator)



from __future__ import annotations

from abc import abstractmethod
from random import choice
from typing import TYPE_CHECKING, Callable, Protocol, Sequence

from yaml import warnings

from negmas.gb.components import GBComponent
from negmas.outcomes.outcome_ops import generalized_minkowski_distance, min_dist
from negmas.preferences import (
    BaseUtilityFunction,
    InverseUFun,
    PresortingInverseUtilityFunction,
    RankOnlyUtilityFunction,
)

if TYPE_CHECKING:
    from negmas.gb import GBState
    from negmas.outcomes import Outcome
    from negmas.outcomes.outcome_space import DistanceFun


__all__ = [
    "OfferSelectorProtocol",
    "OfferSelector",
    "RandomOfferSelector",
    "BestOfferSelector",
    "MedianOfferSelector",
    "WorstOfferSelector",
    "OfferOrientedSelector",
    "FirstOfferOrientedSelector",
    "LastOfferOrientedSelector",
    "BestOfferOrientedSelector",
    "OutcomeSetOrientedSelector",
    "PartnerOffersOrientedSelector",
    "MultiplicativePartnerOffersOrientedSelector",
    "AdditivePartnerOffersOrientedSelector",
]


def additive_score(
    outcomes: Sequence[Outcome],
    partner_offers: Sequence[Outcome],
    ufun: BaseUtilityFunction,
    distance_fun: DistanceFun = generalized_minkowski_distance,
    u_weight: float = 0.5,
    **kwargs,
) -> Sequence[tuple[float, Outcome]]:
    """
    Selects the outcome that maximizes the weightd sum of utility value and distance to the partner offers (with `u_weight` weighing the utility value)

    See Also:

        `min_dist` , `diff`
    """
    utils = [float(ufun(_)) for _ in outcomes]
    dists = [
        min_dist(
            w,
            partner_offers,
            ufun.outcome_space,
            distance_fun=distance_fun,
            **kwargs,
        )
        for w in outcomes
    ]
    max_dist = max(dists)
    if abs(max_dist) < 1e-8:
        scores = sorted(zip(utils, outcomes), reverse=True)
        return scores[0][1]
    dists = [(max_dist - _) / max_dist for _ in dists]
    scores = [
        (u * u_weight + (1 - u_weight) * d, w)
        for (u, d, w) in zip(utils, dists, outcomes)
    ]
    return scores


def multiplicative_score(
    outcomes: Sequence[Outcome],
    partner_offers: Sequence[Outcome],
    ufun: BaseUtilityFunction,
    distance_fun: DistanceFun = generalized_minkowski_distance,
    **kwargs,
) -> Sequence[tuple[float, Outcome]]:
    """
    Selects the outcome that maximizes the product of utility value and distance to the partner offers.

    See Also:

        `min_dist` , `diff`
    """
    utils = [float(ufun(_)) for _ in outcomes]
    dists = [
        min_dist(
            w,
            partner_offers,
            ufun.outcome_space,
            distance_fun=distance_fun,
            **kwargs,
        )
        for w in outcomes
    ]
    max_dist = max(dists)
    if abs(max_dist) < 1e-8:
        scores = sorted(zip(utils, outcomes), reverse=True)
        return scores[0][1]
    dists = [(max_dist - _) / max_dist for _ in dists]
    scores = [(u * d, w) for (u, d, w) in zip(utils, dists, outcomes)]
    return scores


def make_inverter(
    ufun: BaseUtilityFunction,
    ufun_inverter: Callable[[BaseUtilityFunction], InverseUFun] | None = None,
    rank_only: bool = False,
    max_cardinality: int = 10_000,
) -> InverseUFun:
    """
    Creates an `InverseUFun` object from the given ufun with appropriate type if the type is not given

    Args:
        ufun (BaseUtilityFunction): The ufun to invert
        rank_only (bool): If True, only the relative ranks of outcomes will be used in inversion not the values themselves.
        ufun_inverter (Callable[[BaseUtilityFunction], InverseUFun] | None): An optional factory to generate a `InverseUFun` from a `BaseUtilityFunction` .
        max_cardinality (int): The maximum cardinality at which we switch to using a `SamplingInverseUtilityFunction`
    """
    if rank_only:
        ufun = RankOnlyUtilityFunction(ufun, randomize_equal=False, name=ufun.name)
    if ufun_inverter:
        return ufun_inverter(ufun)
    return PresortingInverseUtilityFunction(ufun, max_cache_size=max_cardinality)


class OfferFilterProtocol(Protocol):
    """
    Can select *the best* offers in some  sense from a list of offers based on an inverter
    """

    def __call__(
        self, outcomes: Sequence[Outcome], state: GBState
    ) -> Sequence[Outcome]:
        ...


def NoFiltering(outcomes: Sequence[Outcome], state: GBState) -> Sequence[Outcome]:
    return outcomes


def KeepFirst(outcomes: Sequence[Outcome], state: GBState) -> Sequence[Outcome]:
    return [] if not outcomes else [outcomes[0]]


def KeepLast(outcomes: Sequence[Outcome], state: GBState) -> Sequence[Outcome]:
    return [] if not outcomes else [outcomes[-1]]


class OfferSelectorProtocol(Protocol):
    """
    Can select *the best* offer in some  sense from a list of offers based on an inverter
    """

    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        ...


class OfferSelector(OfferSelectorProtocol, GBComponent):
    """
    Can select *the best* offer in some  sense from a list of offers based on an inverter
    """

    @abstractmethod
    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        ...


class RandomOfferSelector(OfferSelector):
    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not outcomes:
            return None
        return choice(outcomes)


class BestOfferSelector(OfferSelector):
    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not self._negotiator:
            warnings.warn(
                "Asked to select an outcome with unkonwn negotiator",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not self._negotiator.ufun:
            warnings.warn(
                "Asked to select an outcome with unkonwn utility function",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not outcomes:
            return None
        return sorted(((self._negotiator.ufun(_), _) for _ in outcomes), reverse=True)[
            0
        ][1]


class MedianOfferSelector(OfferSelector):
    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not self._negotiator:
            warnings.warn(
                "Asked to select an outcome with unkonwn negotiator",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not self._negotiator.ufun:
            warnings.warn(
                "Asked to select an outcome with unkonwn utility function",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not outcomes:
            return None
        ordered = sorted(
            ((self._negotiator.ufun(_), _) for _ in outcomes), reverse=False
        )
        return ordered[len(ordered) // 2][1]


class WorstOfferSelector(OfferSelector):
    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not self._negotiator:
            warnings.warn(
                "Asked to select an outcome with unkonwn negotiator",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not self._negotiator.ufun:
            warnings.warn(
                "Asked to select an outcome with unkonwn utility function",
                warnings.NegmasUnexpectedValueWarning,
            )
            return None
        if not outcomes:
            return None
        return sorted(((self._negotiator.ufun(_), _) for _ in outcomes), reverse=False)[
            0
        ][1]


class OfferOrientedSelector(OfferSelector):
    """
    Selects the nearest outcome to the pivot outcome which is updated before responding
    """

    def __init__(
        self, distance_fun: DistanceFun = generalized_minkowski_distance, **kwargs
    ):
        self._pivot: Outcome | None = None
        self._distance_fun = distance_fun
        self._distance_fun_params = kwargs

    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not self._negotiator or not self._negotiator.ufun:
            raise ValueError(f"Unknown ufun or negotiator")
        if not self._pivot:
            return choice(outcomes)
        nearest, ndist = None, float("inf")
        for o in outcomes:
            d = self._distance_fun(
                o,
                self._pivot,
                self._negotiator.ufun.outcome_space,
                **self._distance_fun_params,
            )
            if d < ndist:
                nearest, ndist = o, d
        return nearest


class FirstOfferOrientedSelector(OfferOrientedSelector):
    """
    Selects the offer nearest the partner's first offer
    """

    def before_responding(
        self, state: GBState, offer: Outcome | None, source: str
    ) -> Outcome | None:
        if self._pivot or offer is None:
            return
        self._pivot = offer


class LastOfferOrientedSelector(OfferOrientedSelector):
    """
    Selects the offer nearest the partner's last offer
    """

    def before_responding(self, state: GBState, offer: Outcome | None, source: str):
        if offer is None:
            return
        self._pivot = offer


class BestOfferOrientedSelector(OfferOrientedSelector):
    """
    Selects the offer nearest the partner's best offer for me so far
    """

    _pivot_util: float = float("-inf")

    def before_responding(self, state: GBState, offer: Outcome | None, source: str):
        if offer is None:
            return
        if not self._negotiator or not self._negotiator.ufun:
            raise ValueError(f"Unknown ufun or negotiator")
        u = self._negotiator.ufun(offer)
        if u is None:
            return
        u = float(u)
        if u > self._pivot_util:
            self._pivot_util = u
            self._pivot = offer
        self._pivot_util = u


class OutcomeSetOrientedSelector(OfferSelector):
    """
    Selects the nearest outcome to a set of pivot outcomes which is updated before responding
    """

    def __init__(
        self,
        distance_fun: DistanceFun = generalized_minkowski_distance,
        offer_filter: OfferFilterProtocol = NoFiltering,
        **kwargs,
    ):
        self._pivots: list[Outcome] = []
        self._distance_fun = distance_fun
        self._distnace_fun_params = kwargs
        self._offer_filter = offer_filter

    @abstractmethod
    def calculate_scores(
        self, outcomes: Sequence[Outcome], pivots: list[Outcome], state: GBState
    ) -> Sequence[tuple[float, Outcome]]:
        ...

    def __call__(self, outcomes: Sequence[Outcome], state: GBState) -> Outcome | None:
        if not self._negotiator or not self._negotiator.ufun:
            raise ValueError(f"Unknown ufun or negotiator")
        outcomes = self._offer_filter(outcomes, state)
        if not self._pivots:
            return choice(outcomes)
        scores = self.calculate_scores(outcomes, self._pivots, state)
        scores = sorted(
            scores,
            reverse=True,
        )
        return scores[0][1]


class PartnerOffersOrientedSelector(OutcomeSetOrientedSelector):
    """
    Orients offes toward the set of past opponent offers
    """

    def before_responding(self, state: GBState, offer: Outcome | None, source: str):
        if offer is None:
            return
        self._pivots.append(offer)


class MultiplicativePartnerOffersOrientedSelector(PartnerOffersOrientedSelector):
    """
    Orients offes toward the set of past opponent offers.

    The score of an offer is the product of its utility to self and its distance
    to opponent's past offers after normalization
    """

    def calculate_scores(
        self, outcomes: Sequence[Outcome], pivots: list[Outcome], state: GBState
    ) -> Sequence[tuple[float, Outcome]]:
        if not self._negotiator or not self._negotiator.ufun:
            raise ValueError(f"Unknown ufun or negotiator")
        return multiplicative_score(
            outcomes,
            pivots,
            self._negotiator.ufun,
            self._distance_fun,
            **self._distnace_fun_params,
        )


class AdditivePartnerOffersOrientedSelector(PartnerOffersOrientedSelector):
    """
    Orients offes toward the set of past opponent offers.

    The score of an offer is the product of its utility to self and its distance
    to opponent's past offers after normalization
    """

    def __init__(self, *args, u_weight: float = 0.6, **kwargs):
        super().__init__(*args, **kwargs)
        self.u_weight = u_weight

    def calculate_scores(
        self, outcomes: Sequence[Outcome], pivots: list[Outcome], state: GBState
    ) -> Sequence[tuple[float, Outcome]]:
        if not self._negotiator or not self._negotiator.ufun:
            raise ValueError(f"Unknown ufun or negotiator")
        return additive_score(
            outcomes,
            pivots,
            self._negotiator.ufun,
            u_weight=self.u_weight,
            distance_fun=self._distance_fun,
            **self._distnace_fun_params,
        )



from __future__ import annotations

import math
import random
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Iterable

from attrs import define, field

from negmas import warnings
from negmas.common import PreferencesChange, PreferencesChangeType
from negmas.gb.common import ResponseType, current_thread_id

from .base import AcceptancePolicy, FilterResult
from .concession import ConcessionRecommender
from .models.ufun import UFunModel

if TYPE_CHECKING:
    from negmas.common import PreferencesChange
    from negmas.gb import GBState
    from negmas.gb.negotiators.base import GBNegotiator
    from negmas.outcomes import Outcome

    from .offering import MiCROOfferingPolicy, OfferingPolicy

__all__ = [
    "LimitedOutcomesAcceptancePolicy",
    "NegotiatorAcceptancePolicy",
    "ConcensusAcceptancePolicy",
    "AllAcceptanceStrategies",
    "AnyAcceptancePolicy",
    "AcceptImmediately",
    "RejectAlways",
    "EndImmediately",
    "AcceptAbove",
    "RandomAcceptancePolicy",
    "AcceptTop",
    "AcceptBest",
    "TFTAcceptancePolicy",
    "ACNext",
    "ACLast",
    "ACLastKReceived",
    "ACLastFractionReceived",
    "ACTime",
    "AcceptAfter",
    "AcceptAround",
    "AcceptBetween",
    "ACConst",
    "AcceptAnyRational",
    "AcceptBetterRational",
    "AcceptNotWorseRational",
]


@define
class AcceptAnyRational(AcceptancePolicy):
    """
    Accepts any rational outcome.
    """

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.preferences:
            return ResponseType.REJECT_OFFER
        pref = self.negotiator.preferences
        if not pref.is_worse(offer, None):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptNotWorseRational(AcceptancePolicy):
    """
    Accept any outcome not worse than the best so far.
    """

    _accepted: dict[str, Outcome] = field(factory=dict)

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.preferences:
            return ResponseType.REJECT_OFFER
        tid = current_thread_id(state, source)
        current = self._accepted.get(tid, None)
        if offer is None:
            return ResponseType.REJECT_OFFER
        pref = self.negotiator.preferences
        if pref.is_not_worse(offer, current):
            self._accepted[tid] = offer
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptBetterRational(AcceptancePolicy):
    """
    Accept first rational outcomes and then accept only outcomes better than the all accepted so far.
    """

    _accepted: dict[str, Outcome] = field(factory=dict)

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.preferences:
            return ResponseType.REJECT_OFFER
        tid = current_thread_id(state, source)
        current = self._accepted.get(tid, None)
        if offer is None:
            return ResponseType.REJECT_OFFER
        pref = self.negotiator.preferences
        if pref.is_better(offer, current) or (
            current is None and pref.is_not_worse(offer, current)
        ):
            self._accepted[tid] = offer
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACConst(AcceptancePolicy):
    """
    Accepts outcomes with utilities above the given threshold
    """

    th: float = 0.9

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        u = float(self.negotiator.ufun(offer))
        if u >= self.th:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACLastKReceived(AcceptancePolicy):
    """
    Accepts $\\omega$ if $\alpha u(my-next-offer) + \beta > f(u(\text{utils of offers received in the last k steps))$
    """

    k: int = 0
    alpha: float = 1.0
    beta: float = 0.0
    op: Callable[[list[float]], float] = max
    _best: list[float] = field(init=False, default=[])

    def after_join(self, nmi) -> None:
        k = nmi.n_steps if self.k <= 0 and nmi.n_steps else self.k
        self._best = [float("inf") for _ in range(k)]  # type: ignore

    def before_responding(
        self, state: GBState, offer: Outcome | None, source: str | None
    ):
        if not self.negotiator or not self.negotiator.ufun:
            return
        self._best.append(float(self.negotiator.ufun(offer)))
        self._best = self._best[1:]

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        best = self.op(self._best)

        u = float(self.negotiator.ufun(offer))
        if self.alpha * u + self.beta > best:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACLastFractionReceived(AcceptancePolicy):
    """
    Accepts $\\omega$ if $\alpha u(my-next-offer) + \beta > f(u(\text{utils of offers received in the given fraction of time}))$
    """

    fraction: float = 1.0
    alpha: float = 1.0
    beta: float = 0.0
    op: Callable[[list[float]], float] = max
    _best: list[tuple[float, float]] = field(init=False, default=[])

    def before_responding(
        self, state: GBState, offer: Outcome | None, source: str | None
    ):
        if not self.negotiator or not self.negotiator.ufun:
            return
        self._best.append((float(self.negotiator.ufun(offer)), state.relative_time))

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        cutoff = state.relative_time - self.fraction
        lst = []
        for i in range(len(self._best) - 1, -1, -1):
            if self._best[i][1] < cutoff:
                self._best = self._best[i:]
                break
            lst.append(self._best[i][0])
        best = self.op(lst)

        u = float(self.negotiator.ufun(offer))
        if self.alpha * u + self.beta > best:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptFinalOffer(AcceptancePolicy):
    """
    Accepts the final offer as long as it is above the reserved value
    """

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun or not self.negotiator.nmi:
            return ResponseType.REJECT_OFFER
        if (
            self.negotiator.nmi.n_steps == float("inf")
            or self.negotiator.nmi.n_steps is None
        ):
            is_last = state.relative_time > 0.999999
        else:
            is_last = state.step == self.negotiator.nmi.n_steps - 1

        if (
            is_last
            and self.negotiator.ufun(offer) >= self.negotiator.ufun.reserved_value
        ):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACLast(AcceptancePolicy):
    """
    Implements the AClast acceptance strategy based on our last offer.

    Accepts $\\omega$ if $\alpha u(my-next-offer) + \beta > u(\\omega)$
    """

    last_offer_util: float = field(init=False, default=float("inf"))
    alpha: float = 1.0
    beta: float = 0.0

    def after_proposing(self, state: GBState, offer: Outcome | None):
        if not self.negotiator or not self.negotiator.ufun:
            return
        self.last_offer_util = float(self.negotiator.ufun(offer))

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        last = self.last_offer_util
        u = float(self.negotiator.ufun(offer))
        if self.alpha * u + self.beta > last:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptBetween(AcceptancePolicy):
    """
    Accepts in the given range of relative times.
    """

    min: float
    max: float = 1.0

    def __call__(self, state, offer, source):
        if self.max >= state.relative_time >= self.min:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACTime(AcceptancePolicy):
    """
    Implements the ACtime acceptance strategy based on our next offer.

    Accepts if the relative time is greater than or equal to tau
    """

    tau: float

    def __call__(self, state, offer, source):
        if state.relative_time >= self.tau:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


AcceptAfter = ACTime
"""An alias for ACTime"""


@define
class MiCROAcceptancePolicy(AcceptancePolicy):
    """
    Implements the ACnext acceptance strategy based on our next offer.

    Accepts $\\omega$ if $\alpha u(my-next-offer) + \beta > u(\\omega)$
    """

    offering_strategy: MiCROOfferingPolicy
    accept_same: bool = True

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        mine = (
            self.offering_strategy.next_offer()
            if self.offering_strategy.ready_to_concede()
            else self.offering_strategy.best_offer_so_far()
        )
        if self.negotiator.ufun.is_better(None, mine):
            mine = None
        is_acceptable = (
            self.negotiator.ufun.is_not_worse
            if self.accept_same
            else self.negotiator.ufun.is_better
        )
        if is_acceptable(offer, mine):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class ACNext(AcceptancePolicy):
    """
    Implements the ACnext acceptance strategy based on our next offer.

    Accepts $\\omega$ if $\alpha u(my-next-offer) + \beta > u(\\omega)$
    """

    offering_strategy: OfferingPolicy
    alpha: float = 1.0
    beta: float = 0.0

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        next = float(self.negotiator.ufun(self.offering_strategy(state)))
        u = float(self.negotiator.ufun(offer))
        if self.alpha * u + self.beta >= next:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class TFTAcceptancePolicy(AcceptancePolicy):
    """
    An acceptance strategy that concedes as much as the partner (or more)
    """

    partner_ufun: UFunModel
    recommender: ConcessionRecommender

    def __call__(self, state, offer, source):
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        partner_u = float(self.partner_ufun.eval_normalized(offer)) if offer else 1.0
        partner_concession = 1.0 - partner_u
        my_concession = self.recommender(partner_concession, state)
        if not math.isfinite(my_concession):
            warnings.warn(
                f"Got {my_concession} for concession which is unacceptable. Will use no concession"
            )
            my_concession = 0.0
        if not (-1e-6 <= my_concession <= 1.0000001):
            warnings.warn(f"{my_concession} is negative or above 1")
            my_concession = 0.0
        u = 1.0 - float(my_concession)
        if self.negotiator.ufun.eval_normalized(offer) >= u:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptAround(AcceptancePolicy):
    """Accepts around the given relative time (i.e. eps from it)"""

    relative_time: float = 1.0
    eps: float = 0.001

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if offer is None:
            return ResponseType.REJECT_OFFER
        # guarantee acceptance in the last step if relative_time <= 1
        n_steps = self.negotiator.nmi.n_steps
        if self.relative_time <= 1.0 and (
            n_steps is not None and state.step >= n_steps - 1
        ):
            return ResponseType.ACCEPT_OFFER
        # accept around the given time
        if (
            state.relative_time >= self.relative_time
            or (self.relative_time - state.relative_time) < self.eps
        ):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class RandomAcceptancePolicy(AcceptancePolicy):
    p_acceptance: float = 0.15
    p_rejection: float = 0.25
    p_ending: float = 0.1

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        r = random.random()
        if r <= self.p_acceptance + 1e-8:
            return ResponseType.ACCEPT_OFFER
        if r <= self.p_acceptance + self.p_rejection + 1e-8:
            return ResponseType.REJECT_OFFER
        if r <= self.p_acceptance + self.p_rejection + self.p_ending + 1e-8:
            return ResponseType.END_NEGOTIATION
        return ResponseType.NO_RESPONSE


@define
class AcceptBest(AcceptancePolicy):
    """
    Accepts Only the best outcome.

    Remarks:
        - If the best possible utility cannot be found, nothing will be accepted
    """

    _best_util: float = float("inf")

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            return

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER

        if self.negotiator.ufun(offer) >= self.negotiator.ufun.max() - 1e-10:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptTop(AcceptancePolicy):
    """
    Accepts outcomes that are in the given top fraction or top `k`. If neither is given it reverts to accepting the best outcome only.

    Remarks:
        - The outcome-space is always discretized and the constraints `fraction` and `k` are applied to the discretized space
    """

    fraction: float = 0.0
    k: int = 1

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            return
        if any(
            _.type
            not in (
                PreferencesChangeType.Scale,
                PreferencesChangeType.ReservedOutcome,
                PreferencesChangeType.ReservedValue,
            )
            for _ in changes
        ):
            self.negotiator.ufun.invert().init()

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        top_k = self.negotiator.ufun.invert().within_indices((0, self.k))
        if offer in top_k:
            return ResponseType.ACCEPT_OFFER
        top_f = self.negotiator.ufun.invert().within_fractions((0.0, self.fraction))
        if offer in top_f:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class AcceptAbove(AcceptancePolicy):
    """
    Accepts outcomes with utilities in the given top `limit` fraction above reserve/minimum (based on `above_resrve` ).
    """

    limit: float
    above_reserve: bool = True

    def on_preferences_changed(self, changes: list[PreferencesChange]):
        if not self.negotiator or not self.negotiator.ufun:
            return

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if not self.negotiator or not self.negotiator.ufun:
            return ResponseType.REJECT_OFFER
        if (
            self.negotiator.ufun.eval_normalized(offer, self.above_reserve)
            >= self.limit
        ):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class EndImmediately(AcceptancePolicy):
    """
    Rejects immediately anything
    """

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        return ResponseType.END_NEGOTIATION


@define
class RejectAlways(AcceptancePolicy):
    """
    Rejects everything
    """

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        return ResponseType.REJECT_OFFER


@define
class AcceptImmediately(AcceptancePolicy):
    """
    Accepts immediately anything
    """

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        return ResponseType.ACCEPT_OFFER


@define
class LimitedOutcomesAcceptancePolicy(AcceptancePolicy):
    """
    Accepts from a list of predefined outcomes


    Remarks:
        - if `prob` is a number, it is taken as the probability of aceptance for any outcome.
        - if `prob` is `None`, the probability of acceptance of any outcome will be set to the relative time
    """

    prob: dict[Outcome, float] | float | None
    p_ending: float = 0.0

    @classmethod
    def from_outcome_list(
        cls,
        outcomes: list[Outcome],
        prob: list[float] | float = 1.0,
        p_ending: float = 0.0,
    ):
        if not isinstance(prob, Iterable):
            prob = [prob] * len(outcomes)
        return LimitedOutcomesAcceptancePolicy(
            prob=dict(zip(outcomes, prob)), p_ending=p_ending
        )

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        if random.random() < self.p_ending - 1e-12:
            return ResponseType.END_NEGOTIATION
        if self.prob is None:
            prob = state.relative_time
        elif isinstance(self.prob, float):
            prob = self.prob
        else:
            prob = self.prob.get(offer, 0.0)
        if random.random() <= prob:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


@define
class NegotiatorAcceptancePolicy(AcceptancePolicy):
    """
    Uses a negotiator as an offering strategy
    """

    acceptor: GBNegotiator

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        return self.acceptor.respond(state, source)


@define
class ConcensusAcceptancePolicy(AcceptancePolicy, ABC):
    """
    Accepts based on concensus of multiple strategies
    """

    strategies: list[AcceptancePolicy]

    def _set_child_negotiators(self) -> None:
        for s in self.strategies:
            self.set_negotiator(self.negotiator)

    def on_negotiation_start(self, state) -> None:
        self._set_child_negotiators()
        return super().on_negotiation_start(state)

    def filter(self, indx: int, response: ResponseType) -> FilterResult:
        """
        Called with the decision of each strategy in order.


        Remarks:
            - Two decisions need to be made:

              1. Should we continue trying other strategies
              2. Should we save this result.
        """
        return FilterResult(True, True)

    @abstractmethod
    def decide(self, indices: list[int], responses: list[ResponseType]) -> ResponseType:
        """
        Called to make a final decision given the decisions of the strategies with indices `indices` (see `filter` for filtering rules)
        """

    def __call__(
        self, state: GBState, offer: Outcome, source: str | None
    ) -> ResponseType:
        selected, selected_indices = [], []
        for i, s in enumerate(self.strategies):
            s._negotiator = self.negotiator  # type: ignore
            response = s.respond(state, offer, source)
            r = self.filter(i, response)
            if r.save:
                selected.append(response)
                selected_indices.append(i)
            if not r.next:
                break

        return self.decide(selected_indices, selected)


@define
class AllAcceptanceStrategies(ConcensusAcceptancePolicy):
    """Accept only if all children accept, end only if all of them end, otherwise reject"""

    def filter(self, indx: int, response: ResponseType) -> FilterResult:
        if response == ResponseType.REJECT_OFFER:
            return FilterResult(False, True)
        if response == ResponseType.END_NEGOTIATION:
            return FilterResult(False, True)
        return FilterResult(True, False)

    def decide(self, indices: list[int], responses: list[ResponseType]) -> ResponseType:
        if not responses:
            return ResponseType.ACCEPT_OFFER
        return responses[0]


@define
class AnyAcceptancePolicy(ConcensusAcceptancePolicy):
    """Accept any children accept,
    end or reject only if all of them end or reject"""

    def filter(self, indx: int, response: ResponseType) -> FilterResult:
        if response == ResponseType.ACCEPT_OFFER:
            return FilterResult(False, True)
        if response == ResponseType.END_NEGOTIATION:
            return FilterResult(False, True)
        return FilterResult(True, False)

    def decide(self, indices: list[int], responses: list[ResponseType]) -> ResponseType:
        if not responses:
            return ResponseType.REJECT_OFFER
        return ResponseType.ACCEPT_OFFER



# -*- coding: utf-8 -*-
"""
Manages connections to Genius allowing NegMAS users to run Genius agents as negotiators.
"""
from __future__ import annotations

from .bridge import *
from .common import *

# from .gnegotiators import *
from .negotiator import *

# from .ginfo import *


__all__ = (
    common.__all__
    #    + ginfo.__all__
    + bridge.__all__
    + negotiator.__all__
    #    + gnegotiators.__all__
)



from __future__ import annotations

from ..components.acceptance import MiCROAcceptancePolicy
from ..components.offering import MiCROOfferingPolicy
from .modular.mapneg import MAPNegotiator

__all__ = [
    "MiCRONegotiator",
]


class MiCRONegotiator(MAPNegotiator):
    """
    Rational Concession Negotiator

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The ufun of the negotiator (overrides prefrences)
         owner: The `Agent` that owns the negotiator.

    """

    def __init__(self, *args, accept_same: bool = True, **kwargs):
        kwargs["offering"] = MiCROOfferingPolicy()
        kwargs["acceptance"] = MiCROAcceptancePolicy(kwargs["offering"], accept_same)
        super().__init__(*args, **kwargs)



from __future__ import annotations

import os
from pathlib import Path

import hypothesis.strategies as st
import pkg_resources
import pytest
from hypothesis import given, settings

from negmas.genius.bridge import GeniusBridge, genius_bridge_is_running
from negmas.genius.ginfo import get_anac_agents
from negmas.genius.gnegotiators import AgentK, Caduceus
from negmas.genius.negotiator import GeniusNegotiator
from negmas.inout import Scenario
from negmas.sao.mechanism import SAOMechanism
from negmas.sao.negotiators import AspirationNegotiator

TIMELIMIT = 60
STEPLIMIT = 100

SKIP_IF_NO_BRIDGE = not os.environ.get("NEGMAS_LONG_TEST", False)


@given(
    bilateral=st.booleans(),
)
def test_get_genius_agents_example(
    bilateral,
):
    winners = get_anac_agents(bilateral=bilateral, winners_only=True)
    everyone = get_anac_agents(bilateral=bilateral)
    finalists = get_anac_agents(bilateral=bilateral, finalists_only=True)
    for x in (winners, finalists, everyone):
        assert all(
            list(
                isinstance(_, tuple)
                and len(_) == 2
                and isinstance(_[0], str)
                and isinstance(_[1], str)
                and len(_[1]) >= len(_[0])
                for _ in x
            )
        ), x


def test_inclusion_of_sets_in_get_agents():
    from negmas.genius.ginfo import GENIUS_INFO

    for year in GENIUS_INFO.keys():
        winners = get_anac_agents(year=year, winners_only=True)
        finalists = get_anac_agents(year=year, finalists_only=True)
        # everyone = get_anac_agents(year=year)
        assert not finalists or all([_ in finalists for _ in winners]), set(
            winners
        ).difference(set(finalists))
        # assert not everyone or all([_ in everyone for _ in winners]),set(winners).difference(set(everyone))
        # assert not everyone or all([_ in everyone for _ in finalists]), set(finalists).difference(set(everyone))


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_does_not_freeze():
    from pathlib import Path

    from negmas.genius import GeniusNegotiator

    folder_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/cameradomain"
    )
    domain = Scenario.from_genius_folder(Path(folder_name))
    assert domain is not None
    a1 = GeniusNegotiator(
        java_class_name="agents.anac.y2017.ponpokoagent.PonPokoAgent",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[0].name,
    )

    a2 = GeniusNegotiator(
        java_class_name="agents.anac.y2016.yxagent.YXAgent",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[1].name,
    )

    mechanism = domain.make_session(n_steps=None, time_limit=TIMELIMIT)
    mechanism.add(a1)
    mechanism.add(a2)
    mechanism.run()
    # print(a1.preferences.__call__(mechanism.agreement), a2.preferences.__call__(mechanism.agreement))
    GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_old_agent():
    from negmas.genius import GeniusNegotiator

    folder_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/cameradomain"
    )
    domain = Scenario.from_genius_folder(folder_name)
    assert domain is not None
    a1 = GeniusNegotiator(
        java_class_name="agents.anac.y2017.ponpokoagent.PonPokoAgent",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[0].name,
    )

    a2 = GeniusNegotiator(
        java_class_name="agents.anac.y2016.yxagent.YXAgent",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[1].name,
    )

    mechanism = domain.make_session(n_steps=None, time_limit=TIMELIMIT)
    mechanism.add(a1)
    mechanism.add(a2)
    mechanism.run()
    # print(a1.preferences.__call__(mechanism.agreement), a2.preferences.__call__(mechanism.agreement))
    GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_old_agent2():
    from pathlib import Path

    from negmas.genius import GeniusNegotiator

    folder_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/cameradomain"
    )
    domain = Scenario.from_genius_folder(Path(folder_name))
    assert domain is not None
    a1 = GeniusNegotiator(
        java_class_name="agents.anac.y2012.AgentLG.AgentLG",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[0].name,
    )

    a2 = GeniusNegotiator(
        java_class_name="agents.anac.y2016.yxagent.YXAgent",
        domain_file_name=domain.outcome_space.name,
        utility_file_name=domain.ufuns[1].name,
    )

    mechanism = domain.make_session([a1, a2], n_steps=None, time_limit=TIMELIMIT)
    mechanism.run()
    # print(a1.preferences.__call__(mechanism.agreement), a2.preferences.__call__(mechanism.agreement))
    GeniusBridge.clean()


# def test_init_genius_bridge():
#     if not genius_bridge_is_running():
#         init_genius_bridge()
#     assert genius_bridge_is_running()


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
@settings(max_examples=20, deadline=500000)
@given(
    agent_name1=st.sampled_from(GeniusNegotiator.robust_negotiators()),
    agent_name2=st.sampled_from(GeniusNegotiator.robust_negotiators()),
    single_issue=st.booleans(),
)
def test_genius_agents_run_using_hypothesis(
    agent_name1,
    agent_name2,
    single_issue,
):
    src = pkg_resources.resource_filename("negmas", resource_name="tests/data/Laptop")
    base_folder = src
    domain = Scenario.from_genius_folder(Path(base_folder))
    assert domain is not None
    if single_issue:
        domain = domain.to_single_issue()
        assert domain is not None
    a1 = GeniusNegotiator(
        java_class_name=agent_name1,
        preferences=domain.ufuns[0],
    )
    a2 = GeniusNegotiator(
        java_class_name=agent_name2,
        preferences=domain.ufuns[1],
    )
    neg = domain.make_session([a1, a2], n_steps=STEPLIMIT, time_limit=None)
    if neg is None:
        raise ValueError(f"Failed to lead domain from {base_folder}")
    neg._extra_callbacks = True
    neg.run()
    GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agent_gets_preferences():
    base_folder = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    domain = Scenario.from_genius_folder(base_folder)
    assert domain is not None
    a1 = GeniusNegotiator(
        java_class_name="agents.anac.y2015.Atlas3.Atlas3",
        domain_file_name=base_folder + "/Laptop-C-domain.xml",
        utility_file_name=base_folder + f"/Laptop-C-prof1.xml",
    )
    assert a1.preferences is not None
    assert not a1._temp_preferences_file
    assert not a1._temp_domain_file
    a2 = GeniusNegotiator(
        java_class_name="agents.anac.y2015.Atlas3.Atlas3",
        domain_file_name=base_folder + "/Laptop-C-domain.xml",
        preferences=domain.ufuns[0],
    )
    neg = domain.make_session(n_steps=None, time_limit=TIMELIMIT)
    neg.add(a1)
    neg.add(a2)
    assert a2.preferences is not None
    assert a2._temp_preferences_file
    assert not a2._temp_domain_file
    neg.run()
    GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_IF_NO_BRIDGE and not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_genius_agents_run_example():
    from random import randint

    agents = ["agents.anac.y2015.Atlas3.Atlas3", "agents.anac.y2015.AgentX.AgentX"]
    for _ in range(5):
        agent_name1 = agents[randint(0, 1)]
        agent_name2 = agents[randint(0, 1)]
        # print(f"{agent_name1} - {agent_name2}")
        utils = (1, 2)

        base_folder = pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Laptop"
        )
        domain = Scenario.from_genius_folder(base_folder)
        assert domain is not None
        atlas = GeniusNegotiator(
            java_class_name=agent_name1,
            domain_file_name=base_folder + "/Laptop-C-domain.xml",
            utility_file_name=base_folder + f"/Laptop-C-prof{utils[0]}.xml",
        )
        agentx = GeniusNegotiator(
            java_class_name=agent_name2,
            domain_file_name=base_folder + "/Laptop-C-domain.xml",
            utility_file_name=base_folder + f"/Laptop-C-prof{utils[1]}.xml",
        )
        neg = domain.make_session(n_steps=None, time_limit=TIMELIMIT)
        if neg is None:
            raise ValueError(f"Failed to lead domain from {base_folder}")
        neg.add(atlas)
        neg.add(agentx)
        neg.run()

    GeniusBridge.clean()


@pytest.mark.skipif(
    condition=not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_agentk_perceives_time():
    n_steps = 80
    base_folder = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )

    domain = Scenario.from_genius_folder(base_folder)
    assert domain is not None
    gagent = AgentK()
    neg = domain.make_session(
        [gagent, AspirationNegotiator()],
        n_steps=n_steps,
        time_limit=float("inf"),
    )
    if neg is None:
        raise ValueError(f"Failed to load domain from {base_folder}")
    current_time = 0
    for _ in range(n_steps):
        assert gagent.relative_time is not None
        assert (
            gagent.relative_time >= current_time
        ), f"Failed to get time before step {_}"
        neg.step()
        if neg.nmi.state.ended:
            break
        if _ == n_steps - 1:
            assert gagent.relative_time is None, f"Got a time after the last step"
        else:
            assert (
                gagent.relative_time > current_time
            ), f"Failed to get time after step {_}"
        if neg.nmi.state.ended:
            break
        current_time = gagent.relative_time

    assert gagent.relative_time is not None
    assert gagent.relative_time >= current_time


# @pytest.mark.skipif(
#     condition=True or not genius_bridge_is_running(),
#     reason="No Genius Bridge, skipping genius-agent tests",
# )
# def test_running_genius_mechanism_in_genius(tmp_path):
#
#     base_folder = Path(
#         pkg_resources.resource_filename("negmas", resource_name="tests/data/Laptop")
#     )
#     profiles = [
#         "file://" + str(base_folder / "Laptop-C-prof1.xml"),
#         "file://" + str(base_folder / "Laptop-C-prof2.xml"),
#     ]
#     agents = ["agents.anac.y2010.AgentK.Agent_K", "agents.anac.y2010.AgentK.Agent_K"]
#     output_file = str(tmp_path)
#
#     gateway = GeniusBridge.gateway()
#     assert gateway is not None
#     gateway.entry_point.run_negotiation(
#         "genius.core.protocol.StackedAlternatingOffersProtocol",
#         "file://" + str(base_folder / "Laptop-C-domain.xml"),
#         ";".join(profiles),
#         ";".join(agents),
#         output_file,
#     )


@pytest.mark.skipif(
    condition=not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
@pytest.mark.parametrize(
    ["a1", "a2", "n_steps", "time_limit"],
    [
        (
            "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011",
            "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011",
            100,
            None,
        ),
        (
            "agents.anac.y2011.HardHeaded.KLH",
            "agents.anac.y2012.CUHKAgent.CUHKAgent",
            None,
            20,
        ),
        (
            "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011",
            "agents.anac.y2010.Nozomi.Nozomi",
            None,
            30,
        ),
        (
            "agents.TimeDependentAgentBoulware",
            "negotiator.parties.BoulwareNegotiationParty",
            100,
            None,
        ),
    ],
)
def test_2genius_together(a1, a2, n_steps, time_limit):
    base_folder = Path(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Car-A-domain"
        )
    )

    domain = Scenario.from_genius_folder(base_folder)
    assert domain is not None
    neg = domain.make_session(n_steps=n_steps, time_limit=time_limit)
    if neg is None:
        raise ValueError(f"Failed to load domain from {base_folder}")
    neg.add(
        GeniusNegotiator(java_class_name=a1, strict=True, preferences=domain.ufuns[0]),
    )
    neg.add(
        GeniusNegotiator(java_class_name=a2, strict=True, preferences=domain.ufuns[1]),
    )
    neg.run()


@pytest.mark.skipif(
    condition=not genius_bridge_is_running(),
    reason="No Genius Bridge, skipping genius-agent tests",
)
def test_caudacius_caudacius():
    n_steps = 100
    base_folder = Path(
        pkg_resources.resource_filename(
            "negmas", resource_name="tests/data/Car-A-domain"
        )
    )

    domain = Scenario.from_genius_folder(base_folder)
    assert domain is not None
    neg = domain.make_session(n_steps=n_steps, time_limit=float("inf"))
    if neg is None:
        raise ValueError(f"Failed to load domain from {base_folder}")
    if not isinstance(neg, SAOMechanism):
        raise ValueError(f"Loading generated a domain that is not SAO {type(neg)}")
    neg.add(Caduceus(preferences=domain.ufuns[0], strict=True))
    neg.add(Caduceus(preferences=domain.ufuns[1], strict=True))
    for _ in range(n_steps):
        neg.step()
        state: SAOState = neg.state  # type: ignore
        if state.agreement is not None:
            break
        new_offers = [_[1] for _ in state.new_offers]
        assert all(
            _ is not None for _ in new_offers
        ), f"failed at {neg.current_step}: {new_offers}"

    assert not all(
        [len(set(neg.negotiator_offers(_))) == 1 for _ in neg.negotiator_ids]
    ), f"None of the agents conceeded: {neg.trace}"


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations

from abc import abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING

from negmas.preferences.base_ufun import BaseUtilityFunction
from negmas.preferences.preferences import Preferences

from ...negotiators import Controller, Negotiator
from ...outcomes import Outcome
from ..common import GBState, ResponseType, ThreadState

if TYPE_CHECKING:
    from negmas.sao.common import SAOResponse, SAOState
    from negmas.situated import Agent

__all__ = [
    "GBNegotiator",
]


class GBNegotiator(Negotiator):
    """
    Base class for all GB negotiators.

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The utility function of the negotiator (overrides preferences if given)
         owner: The `Agent` that owns the negotiator.

    Remarks:
        - The only method that **must** be implemented by any GBNegotiator is `propose`.
        - The default `respond` method, accepts offers with a utility value no less than whatever `propose` returns
          with the same mechanism state.

    """

    def __init__(
        self,
        preferences: Preferences | None = None,
        ufun: BaseUtilityFunction | None = None,
        name: str | None = None,
        parent: Controller | None = None,
        owner: Agent | None = None,
        id: str | None = None,
        type_name: str | None = None,
        **kwargs,
    ):
        super().__init__(
            name=name,
            preferences=preferences,
            ufun=ufun,
            parent=parent,
            owner=owner,
            id=id,
            type_name=type_name,
            **kwargs,
        )
        self.add_capabilities({"respond": True, "propose": True, "max-proposals": 1})
        self.__end_negotiation = False
        self.__received_offer: dict[str | None, Outcome | None] = defaultdict(
            lambda: None
        )

    @abstractmethod
    def propose(self, state: GBState) -> Outcome | None:
        """Propose an offer or None to refuse.

        Args:
            state: `GBState` giving current state of the negotiation.

        Returns:
            The outcome being proposed or None to refuse to propose

        Remarks:
            - This function guarantees that no agents can propose something with a utility value

        """

    @abstractmethod
    def respond(self, state: GBState, source: str | None) -> ResponseType:
        """Called to respond to an offer. This is the method that should be overriden to provide an acceptance strategy.

        Args:
            state: a `GBState` giving current state of the negotiation.

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).
            - Current offer is accessible through state.threads[source].current_offer as long as source != None otherwise it is None

        """

    def on_partner_proposal(
        self, state: GBState, partner_id: str, offer: Outcome
    ) -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `GBState` giving the state of the negotiation when the offer was porposed.
            partner_id: The ID of the agent who proposed
            offer: The proposal.

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_response(
        self,
        state: GBState,
        partner_id: str,
        outcome: Outcome,
        response: ResponseType,
    ) -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `GBState` giving the state of the negotiation when the partner responded.
            partner_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_ended(self, partner: str):
        """
        Called when a partner ends the negotiation.

        Note  that the negotiator owning this component may never receive this
        offer.
        This is only receivd if the mechanism is sending notifications
        on every offer.
        """

    # compatibility with SAOMechanism
    def __call__(self, state: SAOState) -> SAOResponse:
        """
        Called by the mechanism to counter the offer. It just calls `respond_` and `propose_` as needed.

        Args:
            state: `SAOState` giving current state of the negotiation.
            offer: The offer to be countered. None means no offer and the agent is requested to propose an offer

        Returns:
            Tuple[ResponseType, Outcome]: The response to the given offer with a counter offer if the response is REJECT

        """
        from negmas.sao.common import SAOResponse

        offer = state.current_offer

        if self.__end_negotiation:
            return SAOResponse(ResponseType.END_NEGOTIATION, None)
        if self.preferences is not None:
            changes = self.preferences.changes()
            if changes:
                self.on_preferences_changed(changes)
        if offer is None:
            return SAOResponse(ResponseType.REJECT_OFFER, self.propose_(state=state))
        response = self.respond_(state=state)
        if response == ResponseType.ACCEPT_OFFER:
            return SAOResponse(response, offer)
        if response != ResponseType.REJECT_OFFER:
            return SAOResponse(response, None)
        return SAOResponse(response, self.propose_(state=state))

    def propose_(self, state: SAOState) -> Outcome | None:
        """
        The method directly called by the mechanism (through `counter` ) to ask for a proposal

        Args:
            state: The mechanism state

        Returns:
            An outcome to offer or None to refuse to offer

        Remarks:
            - Depending on the `SAOMechanism` settings, refusing to offer may be interpreted as ending the negotiation
            - The negotiator will only receive this call if it has the 'propose' capability.

        """
        if not self._capabilities["propose"] or self.__end_negotiation:
            return None
        return self.propose(
            state=self._gb_state_from_sao_state(state),
        )

    def respond_(self, state: SAOState) -> ResponseType:
        """The method to be called directly by the mechanism (through `counter` ) to respond to an offer.

        Args:
            state: a `SAOState` giving current state of the negotiation.

        Returns:
            ResponseType: The response to the offer. Possible values are:

                - NO_RESPONSE: refuse to offer. Depending on the mechanism settings this may be interpreted as ending
                               the negotiation.
                - ACCEPT_OFFER: Accepting the offer.
                - REJECT_OFFER: Rejecting the offer. The negotiator will be given the chance to counter this
                                offer through a call of `propose_` later if this was not the last offer to be evaluated
                                by the mechanism.
                - END_NEGOTIATION: End the negotiation
                - WAIT: Instructs the mechanism to wait for this negotiator more. It may lead to cycles so use with care.

        Remarks:
            - The default implementation never ends the negotiation except if an earler end_negotiation notification is
              sent to the negotiator
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).
            - Current offer is accessible through state.threads[source].current_offer as long as source != None otherwise it is None

        """
        offer = state.current_offer
        if self.__end_negotiation:
            return ResponseType.END_NEGOTIATION
        self.__received_offer[state.current_proposer] = offer

        return self.respond(
            state=self._gb_state_from_sao_state(state),
            source=state.current_proposer if state.current_proposer else "",
        )

    def _gb_state_from_sao_state(self, state: SAOState) -> GBState:
        if isinstance(state, GBState):
            return state
        if not self.nmi:
            raise ValueError("No NMI. Cannot convert SAOState to GBState")
        threads = {
            source: ThreadState(
                new_offer=self.__received_offer.get(state.current_proposer, None),
            )
            for source in self.nmi.negotiator_ids
        }

        return GBState(
            running=state.running,
            waiting=state.waiting,
            started=state.started,
            step=state.step,
            time=state.time,
            relative_time=state.relative_time,
            broken=state.broken,
            timedout=state.timedout,
            agreement=state.agreement,
            results=state.results,
            n_negotiators=state.n_negotiators,
            has_error=state.has_error,
            error_details=state.error_details,
            threads=threads,
        )



# -*- coding: utf-8 -*-
r"""Models basic types of utility functions.

Utility functions are at the core of negotiation. Agents engage in negotiations with the goal of maximizing some utility
function. In most cases, these utility functions are assumed to be known a-periori and static for the duration of a
single negotiations.

Notes:
    We try to allow for applications that do not necessary have these two assumptions in the following ways:

    * A utility_function *value* (\ `Value`\ ) can always represent represent a utility_function distribution over all
      possible utility_function values (\ `Distribution`\ ) or a `KnownValue` which is a real number.

    * The base class of all utility_function *functions* is
      `UtilityFunction` and is assumed to map outcomes (\ `Outcome` objects) to the aforementioned generic utility *values*
      (\ `Value` objects).

    * Utility functions can be constructed using any `Callable` which makes it possible to construct them so that
      they change depending on the context or the progression of the negotiation.


"""
from __future__ import annotations
from .base import *
from .mixins import *
from .protocols import *
from .preferences import *
from .base_ufun import *
from .crisp_ufun import *
from .prob_ufun import *
from .inv_ufun import *
from .discounted import *
from .crisp import *
from .prob import *
from .ops import *
from .complex import *
from .value_fun import *

__all__ = (
    base.__all__
    + mixins.__all__
    + protocols.__all__
    + preferences.__all__
    + base_ufun.__all__
    + crisp_ufun.__all__
    + prob_ufun.__all__
    + inv_ufun.__all__
    + discounted.__all__
    + crisp.__all__
    + prob.__all__
    + ops.__all__
    + complex.__all__
    + value_fun.__all__
)



from pathlib import Path

import hypothesis.strategies as st
from hypothesis import HealthCheck, given, settings

from negmas.helpers import unique_name
from negmas.tests.test_situated import DummyWorld, NegAgent, NegPerStepWorld

results = []  # will keep results not to use printing
N_NEG_STEPS = 20


@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(
    single_checkpoint=st.booleans(),
    checkpoint_every=st.integers(0, 6),
    exist_ok=st.booleans(),
)
def test_world_auto_checkpoint(tmp_path, single_checkpoint, checkpoint_every, exist_ok):
    import shutil

    new_folder: Path = tmp_path / unique_name("empty", sep="")
    new_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(new_folder)
    new_folder.mkdir(parents=True, exist_ok=True)
    filename = "mechanism"
    n_steps = 20

    world = DummyWorld(
        n_steps=n_steps,
        checkpoint_every=checkpoint_every,
        checkpoint_folder=new_folder,
        checkpoint_filename=filename,
        extra_checkpoint_info=None,
        exist_ok=exist_ok,
        single_checkpoint=single_checkpoint,
    )

    world.run()

    if 0 < checkpoint_every <= n_steps:
        if single_checkpoint:
            assert (
                len(list(new_folder.glob("*"))) == 2
            ), f"World ran for: {world.current_step}"
        else:
            assert len(list(new_folder.glob("*"))) >= 2 * (
                max(1, world.current_step // checkpoint_every)
            )
    elif checkpoint_every > n_steps:
        assert len(list(new_folder.glob("*"))) == 2
    else:
        assert len(list(new_folder.glob("*"))) == 0


@given(p_request=st.floats(0.0, 1.0))
@settings(deadline=10_000, max_examples=20)
def test_neg_world_steps_normally(p_request):
    n_steps = N_NEG_STEPS
    world = NegPerStepWorld(n_steps)
    for _ in range(5):
        world.join(NegAgent(p_request=p_request, name=f"a{_}"))
    assert world.current_step == 0
    for i in range(n_steps):
        assert world.step()
        assert world.current_step == i + 1
    for i in range(n_steps):
        assert not world.step()
        assert world.current_step == n_steps



"""Defines value functions for single issues"""
from __future__ import annotations

import math
from abc import ABC, abstractmethod
from functools import lru_cache, reduce
from math import cos, e, log, pow, sin
from operator import add
from typing import Any, Callable, Iterable

from attrs import asdict, define

from negmas.helpers.misc import (
    monotonic_minmax,
    monotonic_multi_minmax,
    nonmonotonic_minmax,
    nonmonotonic_multi_minmax,
)
from negmas.helpers.types import is_lambda_function
from negmas.outcomes.base_issue import Issue
from negmas.outcomes.contiguous_issue import ContiguousIssue
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

from .protocols import MultiIssueFun

MAX_CARINALITY = 10_000_000_000

__all__ = [
    "ConstFun",
    "IdentityFun",
    "AffineFun",
    "LinearFun",
    "TriangularFun",
    "LambdaFun",
    "PolynomialFun",
    "QuadraticFun",
    "ExponentialFun",
    "LogFun",
    "SinFun",
    "CosFun",
    "TableFun",
    "TableMultiFun",
    "AffineMultiFun",
    "LinearMultiFun",
    "LambdaMultiFun",
    "make_fun_from_xml",
]


class BaseFun(ABC):
    @property
    def dim(self) -> int:
        return 1

    @abstractmethod
    def minmax(self, input: Issue) -> tuple[float, float]:
        ...

    @classmethod
    def from_dict(cls, d: dict):
        if isinstance(d, cls):
            return d
        _ = d.pop(PYTHON_CLASS_IDENTIFIER, None)
        return cls(**deserialize(d))  # type: ignore Concrete classes will have constructor params

    def to_dict(self) -> dict[str, Any]:
        return serialize(asdict(self))  # type: ignore

    def min(self, input: Issue) -> float:
        mn, _ = self.minmax(input)
        return mn

    def max(self, input: Issue) -> float:
        _, mx = self.minmax(input)
        return mx


@define(frozen=True)
class TableFun(BaseFun):
    mapping: dict

    @lru_cache
    def minmax(self, input: Issue) -> tuple[float, float]:
        return nonmonotonic_minmax(input, self)

    def shift_by(self, offset: float) -> TableFun:
        d = dict()
        for k in self.mapping.keys():
            d[k] = self.mapping[k] + offset
        return TableFun(d)

    def scale_by(self, scale: float) -> TableFun:
        d = dict()
        for k in self.mapping.keys():
            d[k] = self.mapping[k] * scale
        return TableFun(d)

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        issue_name = issue.name
        dtype = "discrete"
        vtype = (
            "integer"
            if issue.is_integer()
            else "real"
            if issue.is_float()
            else "discrete"
        )
        output = f'<issue index="{indx+1}" etype="{dtype}" type="{dtype}" vtype="{vtype}" name="{issue_name}">\n'
        vals = issue.all  # type: ignore
        for i, issue_value in enumerate(vals):
            uu = self(issue_value) + bias
            output += (
                f'    <item index="{i+1}" value="{issue_value}" evaluation="{uu}" />\n'
            )
        output += "</issue>\n"
        return output

    def __call__(self, x) -> float:
        return self.mapping[x]


@define(frozen=True)
class AffineFun(BaseFun):
    slope: float
    bias: float = 0

    @lru_cache
    def minmax(self, input: Issue) -> tuple[float, float]:
        return monotonic_minmax(input, self)

    def shift_by(self, offset: float) -> AffineFun:
        return AffineFun(slope=self.slope, bias=self.bias + offset)

    def scale_by(self, scale: float) -> AffineFun:
        return AffineFun(slope=self.slope * scale, bias=self.bias * scale)

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="linear" parameter0="{bias+self.bias}" parameter1="{self.slope}"></evaluator>\n'
        # elif isinstance(issue, ContiguousIssue) and issue.cardinality > 50_000:
        #     output = f'<issue index="{indx + 1}" etype="real" type="integer" vtype="integer" name="{issue_name}">\n'
        #     output += f'    <evaluator ftype="linear" parameter0="{bias+self.bias}" parameter1="{self.slope}"></evaluator>\n'
        else:
            vals = list(issue.all)
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float) -> float:
        return x * self.slope + self.bias


@define(frozen=True)
class ConstFun(BaseFun):
    bias: float

    @lru_cache
    def minmax(self, input: Issue) -> tuple[float, float]:
        _ = input
        return (self.bias, self.bias)

    def shift_by(self, offset: float) -> ConstFun:
        return ConstFun(bias=offset + self.bias)

    def scale_by(self, scale: float) -> AffineFun:
        return AffineFun(slope=scale, bias=self.bias)

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        return AffineFun(0.0, self.bias).xml(indx, issue, bias)

    def __call__(self, x: float) -> float:
        _ = x
        return self.bias


@define(frozen=True)
class LinearFun(BaseFun):
    slope: float

    @property
    def bias(sef):
        return 0.0

    @lru_cache
    def minmax(self, input: Issue) -> tuple[float, float]:
        return monotonic_minmax(input, self)

    def shift_by(self, offset: float) -> AffineFun:
        return AffineFun(bias=offset, slope=self.slope)

    def scale_by(self, scale: float) -> LinearFun:
        return LinearFun(slope=scale * self.slope)

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        return AffineFun(self.slope, 0.0).xml(indx, issue, bias)

    def __call__(self, x: float) -> float:
        return x * self.slope


@define(frozen=True)
class IdentityFun(BaseFun):
    def minmax(self, input: Issue) -> tuple[float, float]:
        return (input.min_value, input.max_value)

    def shift_by(self, offset: float) -> ConstFun:
        return ConstFun(bias=offset)

    def scale_by(self, scale: float) -> LinearFun:
        return LinearFun(slope=scale)

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        return LinearFun(1.0).xml(indx, issue, bias)

    def __call__(self, x: float) -> float:
        return x


@define(frozen=True)
class LambdaFun(BaseFun):
    f: Callable[[Any], float]
    bias: float = 0
    min_value: float | None = None
    max_value: float | None = None

    def __post_init__(self):
        # we need to be sure that f is a lambda function so that it can
        # correctly be serialized
        if not is_lambda_function(self.f):
            f = self.f
            self.f = lambda x: f(x)

    def minmax(self, input) -> tuple[float, float]:
        if self.min_value is not None and self.max_value is not None:
            return self.min_value, self.max_value
        mn, mx = nonmonotonic_minmax(input, self.f)
        if self.min_value is not None:
            mn = min(mn, self.min_value)
        if self.max_value is not None:
            mx = min(mx, self.max_value)
        return mn, mx

    def shift_by(self, offset: float, change_bias: bool = False) -> LambdaFun:
        mn, mx = self.min_value, self.max_value
        return LambdaFun(
            self.f if change_bias else lambda x: offset + self.f(x),
            bias=(self.bias + offset) if change_bias else self.bias,
            min_value=mn if mn is None else mn + offset,
            max_value=mx if mx is None else mx + offset,
        )

    def scale_by(self, scale: float) -> LambdaFun:
        mn, mx = self.min_value, self.max_value
        if scale < 0:
            mn, mx = mx, mn
        return LambdaFun(
            lambda x: scale * self.f(x),
            bias=self.bias * scale,
            min_value=mn if mn is None else mn * scale,
            max_value=mx if mx is None else mx * scale,
        )

    def xml(self, indx: int, issue: Issue, bias=0.0) -> str:
        if issue.is_discrete():
            values = list(issue.all)  # type: ignore (I know it is discrete)
            return TableFun(dict(zip(values, [self(_) for _ in values]))).xml(
                indx, issue, bias
            )
        raise ValueError(
            f"LambdaFun with a continuous issue cannot be converted to XML"
        )

    def __call__(self, x: Any) -> float:
        return self.f(x) + self.bias


@define(frozen=True)
class QuadraticFun(BaseFun):
    a2: float
    a1: float
    bias: float = 0

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        mn, mx = input.min_value, input.max_value
        middle = -self.a1 / (2 * self.a2)
        fmn, fmx = self(mn), self(mx)
        if middle < mn or middle > mx:
            if fmn < fmx:
                return fmn, fmx
            return fmx, fmn
        if fmn > fmx:
            fmn = fmx
        fmiddle = self(middle)  # type: ignore
        if fmn < fmiddle:
            return fmn, fmiddle
        return fmiddle, fmn

    def shift_by(self, offset: float) -> QuadraticFun:
        return QuadraticFun(bias=self.bias + offset, a1=self.a1, a2=self.a2)

    def scale_by(self, scale: float) -> QuadraticFun:
        return QuadraticFun(
            bias=self.bias * scale, a1=self.a1 * scale, a2=self.a2 * scale
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="quadratic" parameter0="{bias+self.bias}" parameter1="{self.a1} parameter2={self.a2}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="quadratic" parameter0="{bias+self.bias}" parameter1="{self.a1} parameter2={self.a2}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return self.a2 * x * x + self.a1 * x + self.bias


@define(frozen=True)
class PolynomialFun(BaseFun):
    coefficients: tuple[float]
    bias: float = 0

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        # todo: implement this exactly without sampling
        return nonmonotonic_minmax(input, self)

    def shift_by(self, offset: float) -> PolynomialFun:
        return PolynomialFun(bias=self.bias + offset, coefficients=self.coefficients)

    def scale_by(self, scale: float) -> PolynomialFun:
        return PolynomialFun(
            bias=self.bias * scale,
            coefficients=tuple(_ * scale for _ in self.coefficients),
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="poynomial" parameter0="{bias+self.bias}"'
            for i, x in enumerate(self.coefficients):
                output += f'parameter{i}="{x}"'

            output += "></evaluator>\n"
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="poynomial" parameter0="{bias+self.bias}"'
            for i, x in enumerate(self.coefficients):
                output += f'parameter{i}="{x}"'

            output += "></evaluator>\n"
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return reduce(
            add, [b * pow(x, p + 1) for p, b in enumerate(self.coefficients)], self.bias
        )


@define(frozen=True)
class TriangularFun(BaseFun):
    start: float
    middle: float
    end: float
    bias: float = 0

    def shift_by(self, offset: float) -> TriangularFun:
        return TriangularFun(
            bias=self.bias + offset, start=self.start, middle=self.middle, end=self.end
        )

    def scale_by(self, scale: float) -> TriangularFun:
        return TriangularFun(
            bias=self.bias * scale,
            start=self.start * scale,
            middle=self.middle * scale,
            end=self.end * scale,
        )

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        # todo: implement this exactly without sampling
        return nonmonotonic_minmax(input, self)

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            assert abs(bias + self.bias) < 1e-6
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="triangular" parameter0="{self.start}" parameter1="{self.end} parameter2={self.middle}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            assert abs(bias + self.bias) < 1e-6
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="triangular" parameter0="{self.start}" parameter1="{self.end} parameter2={self.middle}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        bias1, slope1 = self.start, (self.middle - self.start)
        bias2, slope2 = self.middle, (self.middle - self.end)
        return self.bias + (
            bias1 + slope1 * float(x) if x < self.middle else bias2 + slope2 * float(x)
        )


@define(frozen=True)
class ExponentialFun(BaseFun):
    tau: float
    bias: float = 0
    base: float = e

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        return monotonic_minmax(input, self)

    def shift_by(self, offset: float) -> ExponentialFun:
        return ExponentialFun(bias=self.bias + offset, tau=self.tau)

    def scale_by(self, scale: float) -> ExponentialFun:
        return ExponentialFun(
            bias=self.bias * scale, tau=self.tau + math.log(scale, base=self.base)
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="exponential" parameter0="{bias+self.bias}" parameter1="{self.tau} parameter2={self.base}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="exponential" parameter0="{bias+self.bias}" parameter1="{self.tau} parameter2={self.base}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return pow(self.base, self.tau * x) + self.bias


@define(frozen=True)
class CosFun(BaseFun):
    multiplier: float = 1.0
    bias: float = 0.0
    phase: float = 0.0
    amplitude: float = 1.0

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        # todo: implement this exactly without sampling
        return nonmonotonic_minmax(input, self)

    def shift_by(self, offset: float) -> CosFun:
        return CosFun(
            bias=self.bias + offset,
            multiplier=self.multiplier,
            phase=self.phase,
            amplitude=self.amplitude,
        )

    def scale_by(self, scale: float) -> CosFun:
        return CosFun(
            amplitude=self.amplitude * scale,
            bias=self.bias * scale,
            multiplier=self.multiplier,
            phase=self.phase,
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="cos" parameter0="{bias+self.bias}" parameter1="{self.amplitude} parameter2={self.multiplier} parameter3={self.phase}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="cos" parameter0="{bias+self.bias}" parameter1="{self.amplitude} parameter2={self.multiplier} parameter3={self.phase}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return self.amplitude * (cos(self.multiplier * x + self.phase)) + self.bias


@define(frozen=True)
class SinFun(BaseFun):
    multiplier: float = 1.0
    bias: float = 0.0
    phase: float = 0.0
    amplitude: float = 1.0

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        # todo: implement this exactly without sampling
        return nonmonotonic_minmax(input, self)

    def shift_by(self, offset: float) -> SinFun:
        return SinFun(
            bias=self.bias + offset,
            multiplier=self.multiplier,
            phase=self.phase,
            amplitude=self.amplitude,
        )

    def scale_by(self, scale: float) -> SinFun:
        return SinFun(
            amplitude=self.amplitude * scale,
            bias=self.bias * scale,
            multiplier=self.multiplier,
            phase=self.phase,
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="sin" parameter0="{bias+self.bias}" parameter1="{self.amplitude} parameter2={self.multiplier} parameter3={self.phase}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="sin" parameter0="{bias+self.bias}" parameter1="{self.amplitude} parameter2={self.multiplier} parameter3={self.phase}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return self.amplitude * (sin(self.multiplier * x + self.phase)) + self.bias


@define(frozen=True)
class LogFun(BaseFun):
    tau: float
    bias: float = 0
    base: float = e
    scale: float = 1.0

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        return monotonic_minmax(input, self)

    def shift_by(self, offset: float) -> LogFun:
        return LogFun(
            bias=self.bias + offset, tau=self.tau, scale=self.scale, base=self.base
        )

    def scale_by(self, scale: float) -> LogFun:
        return LogFun(
            bias=self.bias * scale,
            tau=self.tau,
            scale=self.scale * scale,
            base=self.base,
        )

    def xml(self, indx: int, issue: Issue, bias) -> str:
        issue_name = issue.name
        if issue.is_continuous():
            output = f'<issue index="{indx + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
            output += f'    <evaluator ftype="log" parameter0="{bias+self.bias}" parameter1="{self.tau} parameter2={self.base} paramter3={self.scale}"></evaluator>\n'
        elif isinstance(issue, ContiguousIssue):
            output = f'<issue index="{indx + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            output += f'    <evaluator ftype="log" parameter0="{bias+self.bias}" parameter1="{self.tau} parameter2={self.base} parameter3={self.scale}"></evaluator>\n'
        else:
            vals = list(issue.all)  # type: ignore
            return TableFun(dict(zip(vals, [self(_) for _ in vals]))).xml(
                indx, issue, bias
            )
        output += "</issue>\n"
        return output

    def __call__(self, x: float):
        return self.scale * log(self.tau * x, base=self.base) + self.bias


@define(frozen=True)
class TableMultiFun(MultiIssueFun):
    mapping: dict[tuple, Any]

    @lru_cache
    def minmax(self, input: Iterable[Issue]) -> tuple[float, float]:
        return nonmonotonic_multi_minmax(input, self)

    def shift_by(self, offset: float) -> TableMultiFun:
        d = dict()
        for k in self.mapping.keys():
            d[k] = self.mapping[k] + offset
        return TableMultiFun(d)

    def scale_by(self, scale: float) -> TableMultiFun:
        d = dict()
        for k in self.mapping.keys():
            d[k] = self.mapping[k] * scale
        return TableMultiFun(d)

    def dim(self) -> int:
        if not len(self.mapping):
            raise ValueError("Unkonwn dictionary in TableMultiFun")
        return len(list(self.mapping.keys())[0])

    def xml(self, indx: int, issues: list[Issue], bias=0) -> str:
        raise NotImplementedError()

    def __call__(self, x):
        return self.mapping[x]


@define(frozen=True)
class AffineMultiFun(MultiIssueFun):
    slope: tuple[float, ...]
    bias: float = 0

    @lru_cache
    def minmax(self, input: Iterable[Issue]) -> tuple[float, float]:
        return monotonic_multi_minmax(input, self)

    def shift_by(self, offset: float) -> AffineMultiFun:
        return AffineMultiFun(slope=self.slope, bias=self.bias + offset)

    def scale_by(self, scale: float) -> AffineMultiFun:
        return AffineMultiFun(
            slope=tuple(scale * _ for _ in self.slope), bias=self.bias * scale
        )

    def dim(self) -> int:
        raise NotImplementedError()

    def xml(self, indx: int, issues: list[Issue], bias=0) -> str:
        raise NotImplementedError()

    def __call__(self, x: tuple):
        return reduce(add, [a * b for a, b in zip(self.slope, x)], self.bias)


@define(frozen=True)
class LinearMultiFun(MultiIssueFun):
    slope: tuple[float, ...]

    @property
    def bias(self):
        return 0

    @lru_cache
    def minmax(self, input: Iterable[Issue]) -> tuple[float, float]:
        return monotonic_multi_minmax(input, self)

    def shift_by(self, offset: float) -> AffineMultiFun:
        return AffineMultiFun(slope=self.slope, bias=self.bias + offset)

    def scale_by(self, scale: float) -> LinearMultiFun:
        return LinearMultiFun(slope=tuple(scale * _ for _ in self.slope))

    def dim(self) -> int:
        raise NotImplementedError()

    def xml(self, indx: int, issues: list[Issue], bias=0) -> str:
        raise NotImplementedError()

    def __call__(self, x: tuple):
        return reduce(add, [a * b for a, b in zip(self.slope, x)], 0)


@define(frozen=True)
class LambdaMultiFun(MultiIssueFun):
    f: Callable[[Any], float]
    bias: float = 0
    min_value: float | None = None
    max_value: float | None = None

    @lru_cache
    def minmax(self, input) -> tuple[float, float]:
        if self.min_value is not None and self.max_value is not None:
            return self.min_value, self.max_value
        mn, mx = nonmonotonic_multi_minmax(input, self.f)
        if self.min_value is not None:
            mn = min(mn, self.min_value)
        if self.max_value is not None:
            mx = min(mx, self.max_value)
        return mn, mx

    def shift_by(self, offset: float) -> AffineMultiFun:
        raise NotImplementedError()

    def scale_by(self, scale: float) -> LinearMultiFun:
        raise NotImplementedError()

    def dim(self) -> int:
        raise NotImplementedError()

    def xml(self, indx: int, issues: list[Issue], bias=0) -> str:
        raise NotImplementedError()

    def __call__(self, x: Any) -> float:
        return self.f(x) + self.bias


def make_fun_from_xml(item) -> tuple[BaseFun, str]:
    if item.attrib["ftype"] == "linear":
        offset = item.attrib.get(
            "offset",
            item.attrib.get("parameter0", item.attrib.get("offset", 0.0)),
        )
        slope = item.attrib.get(
            "slope",
            item.attrib.get("parameter1", item.attrib.get("slope", 1.0)),
        )
        offset, slope = float(offset), float(slope)
        if offset != 0:
            return AffineFun(bias=offset, slope=slope), "affine"
        else:
            return LinearFun(slope=slope), "linear"
    elif item.attrib["ftype"] == "quadratic":
        offset = float(item.attrib.get("parameter0", item.attrib.get("offset", 0.0)))
        a1 = float(item.attrib.get("parameter1", item.attrib.get("a1", 1.0)))
        a2 = float(item.attrib.get("parameter2", item.attrib.get("a2", 1.0)))
        return QuadraticFun(a1=a1, a2=a2, bias=offset), "quadratic"
    elif item.attrib["ftype"] == "triangular":
        strt = float(item.attrib.get("parameter0", item.attrib.get("start", 0.0)))
        end = float(item.attrib.get("parameter1", item.attrib.get("end", 1.0)))
        middle = float(item.attrib.get("parameter2", item.attrib.get("middle", 1.0)))
        return TriangularFun(start=strt, end=end, middle=middle), "triangular"
    else:
        # todo: implement all other functions defined in value_fun.py
        raise ValueError(f'Unknown ftype {item.attrib["ftype"]}')
